/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_canvas.h 
 * @brief Generic canvas main class.
 * 
 */

#ifndef __GC_CANVAS_H__
#define __GC_CANVAS_H__

#include "myx_gc_datatypes.h"
#include "myx_gc_utilities.h"
#include "myx_gc_layer.h"
#include "myx_gc_view.h"

//----------------------------------------------------------------------------------------------------------------------

class CFigure;
class CFigureInstance;
class CLayer;
class CGenericCanvas;
class CGCView;

//----------------- General listeners ----------------------------------------------------------------------------------

class CGCListener;

typedef vector<CGCListener*> CListeners;
typedef vector<CGCListener*>::iterator CListenerIterator;

/**
 * The general listener class is used to notify users of the canvas about general events like repaints and errors.
 * This class is only an abstract class and must get a concrete implemention in the application.
 * All Listener classes are meant to be a means for calling back the application. They are implemented and instantiated
 * in the application and must be freed there. Don't forget to remove the listener class before you free it!
 */
class GENERIC_CANVAS_API CGCListener
{
public:
  virtual void __cdecl OnChange(void* AObject, TGCChangeReason Reason) = 0; // Called for changes in the canvas.
  virtual void __cdecl OnError(const char* Message) = 0;  // Called by the canvas when an error occured.
  virtual void __cdecl OnInvalidate(void) = 0;             // Called by the canvas when the viewer needs a refresh.
};

//----------------- CFigureInstanceEnumerator --------------------------------------------------------------------------

/**
 * The CFigureInstanceEnumerator class is for quick access to all figure instances on all (common) layers.
 * Enumeration happens depth-first. That means for each layer first all instances are enumerated before the next layer
 * is taken.
 */
class GENERIC_CANVAS_API CFigureInstanceEnumerator
{
private:
  CGenericCanvas* FCanvas;
  CLayers::iterator FLayerIterator;
  CFigureInstances::iterator FFigureInstanceIterator;
public:
  CFigureInstanceEnumerator(CGenericCanvas* Canvas);

  virtual bool __cdecl HasNext(void);
  virtual CFigureInstance* __cdecl Next(void);
  virtual void __cdecl Release(void);
  virtual void __cdecl Reset(void);
};

//----------------- CGenericCanvas -------------------------------------------------------------------------------------

#define GC_STATE_PENDING_ACTIVATION 0x0001

/**
 * CGenericCanvas is the main class of the library and is the base for all further functionality (e.g. it creates and 
 * maintains the model). Instances are created via the exported CreateGenericCanvas function (if called from non C++ 
 * languages). CGenericCanvas serves as the controller in the model-view-controller pattern, which is used here and 
 * communicates with the viewer via callbacks.
 * The viewer is platform specific and must be implemented individually. It is responsible to create a canvas controller class.
 *
 * @see CreateGenericCanvas
 */
class GENERIC_CANVAS_API CGenericCanvas 
{
  friend class CGCModel;
  friend class CFeedbackLayer;
  friend class CFigureInstanceEnumerator;
private:
  GCContext FContext;                  // The always active OpenGL rendering cntext.
  CGCModel* FModel;                    // The model this canvas is controlling.
  CLayers FLayers;                     // A list of layers currently in this canvas.
  CFeedbackLayer* FFeedbackLayer;      // The selection layer is special.
  int FUpdateCount;                    // If > 0 then the canvas is currently updating internal structures. Display will not update.
  CViews FViews;                       // A list of views that can be used by this canvas.
  CGCView* FCurrentView;               // The currently active set of layers.
  unsigned int FStates;               // Persistent storage for certain states in the canvas.

  // Listeners
  CListeners FListeners;               // A list of listeners, which want to get notified about general things happening in the canvas.
protected:
  void ClearBuffers();
public:
  CGenericCanvas(GCContext Context);
  virtual ~CGenericCanvas(void);

  virtual void __cdecl AddLayer(CLayer* Layer);
  virtual TGCError __cdecl AddLayoutsFromFile(const char* FileName);
  virtual void __cdecl AddListener(CGCListener* Listener);
  int AddNewItem(string Path, string Name);
  virtual TGCError __cdecl AddStylesFromFile(const char* FileName);
  virtual void __cdecl AddToSelection(CFigureInstance* Instance);
  virtual void __cdecl BeginUpdate(void);
  virtual void __cdecl Change(void* AObject, TGCChangeReason Reason);
  virtual void __cdecl CheckError(void);
  int ChildCount(string Path);
  virtual void __cdecl ClearContent(void);
  virtual void __cdecl ClearLayouts(void);
  virtual void __cdecl ClearSelection(void);
  virtual void __cdecl ClearStyles(void);
  virtual CFigure* __cdecl CreateFigure(const char* Type);
  virtual CLayer* __cdecl CreateLayer(const char* Name, TGCLayerType Type, bool AddToCurrentView);
  virtual CGCView* __cdecl CreateView(const char* Name);
  virtual void __cdecl CurrentView(CGCView* View);
  virtual CGCView* __cdecl CurrentView(void);
  virtual void __cdecl EndUpdate(void);
  virtual void __cdecl Error(const char* Message);
  virtual CFigureInstanceEnumerator* __cdecl GetFigureInstanceEnumerator(void);
  virtual TGCDirection __cdecl GetSelectionInfo(const int X, const int Y);
  virtual void __cdecl Invalidate(void);
  virtual void __cdecl InvalidateSelectionBounds(CFigureInstance* Instance);
  virtual bool __cdecl IsUpdating(void);
  virtual CLayer* __cdecl LayerByName(const char* Name);
  TGCVariant Property(string Path);
  void Property(string Path, const TGCVariant& Value);
  virtual void __cdecl Release(void);
  virtual void __cdecl RemoveFromSelection(CFigureInstance* Instance);
  void RemoveItem(string Path, int Index);
  virtual void __cdecl RemoveLayer(CLayer* Layer);
  virtual void __cdecl RemoveListener(CGCListener* Listener);
  virtual void __cdecl RemoveView(CGCView* View);
  virtual void __cdecl Render(void);
  virtual void __cdecl ResizeFiguresStart(int X, int Y, TGCDirection Direction);
  virtual void __cdecl ResizeFiguresStop(void);
  virtual void __cdecl ResizeFiguresTo(int X, int Y);
  virtual void __cdecl RubberbandResize(int X, int Y, TRBSelectionAction Action);
  virtual void __cdecl RubberbandStart(TRubberbandStyle Style, int X, int Y, bool ClearSelection);
  virtual void __cdecl RubberbandStop(void);
  virtual void __cdecl ShowSelection(bool Visible);
  virtual CGCView* __cdecl ViewByName(const char* Name);
};

//----------------------------------------------------------------------------------------------------------------------

// Factory function to create a generic canvas. This function is exported and must be used by the viewer implementations
// to actually create a canvas instance. This is the only way to get hold of a generic canvas instance for non-C++ languages.
extern "C" GENERIC_CANVAS_API CGenericCanvas* CreateGenericCanvas(GCContext Context);

#endif // __GC_CANVAS_H__
