/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_view.h 
 * @brief Implementation of the view class.
 * 
 */

#ifndef __GC_VIEW_H__
#define __GC_VIEW_H__

#include "myx_gc_datatypes.h"
#include "myx_gc_layer.h"

//----------------- Hit testing structures -----------------------------------------------------------------------------

typedef struct tagHitEntry
{
  CFigureInstance* Instance;
  double ZMin;
  double ZMax;
} THitEntry;

typedef map<CFigureInstance*, THitEntry*> THitEntries;
typedef map<CFigureInstance*, THitEntry*>::iterator THitEntryIterator;

/**
 * The CHitResult class is used to collect a number of figures that are located at a given point in the canvas.
 */
class GENERIC_CANVAS_API CHitResults         
{
  friend class CGCView;
  friend class CLayer;
private:
  THitEntries FEntries;
  THitEntryIterator FCurrentEntry;
protected:
  void AddHit(CFigureInstance* Instance, double Min, double Max);
public:
  CHitResults(void);
  virtual ~CHitResults(void);

  virtual int __cdecl Count(void);
  virtual bool __cdecl Delete(CFigureInstance* Instance);
  virtual void __cdecl DeleteCurrent(void);
  virtual bool __cdecl HasNext(void);
  virtual THitEntry* __cdecl Next(void);
  virtual void __cdecl Release(void);
  virtual void __cdecl Reset(void);
};

//----------------------------------------------------------------------------------------------------------------------

class CGenericCanvas;


/** 
 * A view implements an association between a set of layers and their visual representation on screen.
 * Views can have individual zoom and offset values, viewports and other properties.
 * There can always be only one active view. Views are managed by the canvas class.
 */
class CGCView
{
  friend class CGenericCanvas;
private:
  CGenericCanvas* FCanvas;   // The owning canvas.
  string FName;              // An identifier for this view (UTF-8 encoded).
  GLfloat FColor[4];         // The background color of the viewer.
  TGCViewport FViewport;     // Coordinates of the output area within the viewer window.
  CLayers FLayers;           // A list of layers that are visible in this view.
  float FZoomX;              // The current horizontal zoom factor. It's a normal scaling factor but for the 
                             // whole view (scales everything).
  float FZoomY;              // The current vertical zoom factor.
  float FOffsetX;            // A translation offset to enable scrolling (panning) in a zoomed scene.
  float FOffsetY;            // The vertical translation offset.
  float FNearPlane;          // The cut off plane that is closest to the eye.
  float FFarPlane;           // The cut off plane that is farthest from the eye.
  float FJitter;             // Little offset to add to the viewport values to improve display quality on some boards.
  bool FIsPicking;           // true if the model is currently picking figures.
protected:
  void Activate(void);
  void ApplyTransformations(bool DoProjection);
  void Invalidate(void);
  void Render(void);
public:
  CGCView(CGenericCanvas* Owner, string Name);
  virtual ~CGCView() {};

  virtual void __cdecl AddLayer(CLayer* Layer);
  int AddNewItem(string Path, string Name);
  int ChildCount(string Path);
  void Change(void* AObject, TGCChangeReason Reason);
  virtual void __cdecl Color(GLfloat* NewColor);
  virtual void __cdecl Color(float Red, float Green, float Blue, float Alpha);
  virtual bool __cdecl Contains(CLayer* Layer);
  virtual CHitResults* __cdecl GetHitTestInfoAt(const int X, const int Y);
  virtual void __cdecl Jitter(float Value);
  virtual float __cdecl Jitter(void);
  virtual void __cdecl OffsetX(float Value);
  virtual float __cdecl OffsetX(void);
  virtual void __cdecl OffsetY(float Value);
  virtual float __cdecl OffsetY(void);
  TGCVariant Property(string Path);
  void Property(string Path, const TGCVariant& Value);
  void RemoveItem(string Path, int Index);
  virtual void __cdecl RemoveLayer(CLayer* Layer);
  virtual void __cdecl Viewport(const TGCViewport& NewViewport);
  virtual TGCViewport __cdecl Viewport(void);
  virtual void __cdecl ZoomX(float Value);
  virtual float __cdecl ZoomX(void);
  virtual void __cdecl ZoomY(float Value);
  virtual float __cdecl ZoomY(void);
};

//----------------------------------------------------------------------------------------------------------------------

#endif // __GC_VIEW_H__
