/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_canvas.cpp 
 * @brief Generic canvas main class and entry point.
 * 
 */

#include "myx_gc.h"
#include "myx_gc_figure.h"
#include "myx_gc_layer.h"
#include "myx_gc_model.h"
#include "myx_gc_gl_helper.h"
#include "myx_gc_font_manager.h"
#include "myx_gc_texture.h"
#include "myx_gc_svgparser.h"
#include "myx_gc_figure_parser.h"
#include "myx_gc_canvas.h"


//----------------------------------------------------------------------------------------------------------------------

// Factory function for a canvas.
GENERIC_CANVAS_API CGenericCanvas* CreateGenericCanvas(GCContext Context)
{
	return new CGenericCanvas(Context);
}

//----------------- CFigureInstanceEnumerator --------------------------------------------------------------------------

/**
 * Constructor of the enumerator class.
 *
 * @param Canvas The canvas which contains the layers which are to be enumerated.
 */
CFigureInstanceEnumerator::CFigureInstanceEnumerator(CGenericCanvas* Canvas)
{
  FCanvas = Canvas;
  FLayerIterator = FCanvas->FLayers.end();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines if there is a next figure instance to enumerate.
 *
 * @return True if there is still a figure instance otherwise false.
 */
bool CFigureInstanceEnumerator::HasNext(void)
{
  return FLayerIterator != FCanvas->FLayers.end();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the next figure instance in the sequence.
 *
 * @return The next figure instance.
 */
CFigureInstance* CFigureInstanceEnumerator::Next(void)
{
  CFigureInstance* Result = *FFigureInstanceIterator;

  // Advance to next instance.
  ++FFigureInstanceIterator;
  while (true)
  {
    CLayer* Layer = *FLayerIterator;
    if (FFigureInstanceIterator != Layer->FInstances.end())
      break;

    // This layer is exhausted. Get the next one.
    ++FLayerIterator;
    if (FLayerIterator == FCanvas->FLayers.end())
    {
      // No more layers.
      break;
    };

    if (!(*FLayerIterator)->FEnabled)
      FFigureInstanceIterator = (*FLayerIterator)->FInstances.end();
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Frees this enumerator instance. Usually called by non-C++ languages as memory is managed by the C++ runtime.
 */
void CFigureInstanceEnumerator::Release(void)
{
  delete this;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Resets the enumerator to the first figure instance in the canvas.
 */
void CFigureInstanceEnumerator::Reset(void)
{
  FLayerIterator = FCanvas->FLayers.begin();
  while (FLayerIterator != FCanvas->FLayers.end())
  {
    if ((*FLayerIterator)->FEnabled)
    {
      CLayer* Layer = *FLayerIterator;
      FFigureInstanceIterator = Layer->FInstances.begin();
      if (FFigureInstanceIterator == Layer->FInstances.end())
      {
        ++FLayerIterator;
      }
      else
        break; // Found the first instance.
    }
    else
      ++FLayerIterator;
  };
}

//----------------- CGenericCanvas -------------------------------------------------------------------------------------

CGenericCanvas::CGenericCanvas(GCContext Context)
{ 
  FContext = Context;
  FUpdateCount = 0;
  FCurrentView = NULL;
  FStates = 0;

  FModel = new CGCModel(this);
  FFeedbackLayer = new CFeedbackLayer("Feedback");
  FFeedbackLayer->Canvas(this);

  // Set a new lock on the font manager (and create it by the way if not yet done).
  LockFontManager();
}

//----------------------------------------------------------------------------------------------------------------------

CGenericCanvas::~CGenericCanvas(void)
{
  // Free any dynamically allocated memory. Explicitely increase update count. We don't want any recursive
  // notifcations anymore.
  ++FUpdateCount;
  FFeedbackLayer->BeginUpdate();
  for (CLayers::iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    delete *Iterator;

  delete FModel;
  delete FFeedbackLayer;

  // Release the lock we placed in the constructor.
  // If this is the last GC instance then also the font manager can be released.
  UnlockFontManager();
}

//----------------------------------------------------------------------------------------------------------------------

void CGenericCanvas::ClearBuffers(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Adds the given layer at the end of the layer list. The new layer becomes the top layer in the view then.
 * @param Layer The layer to add.
 */
void CGenericCanvas::AddLayer(CLayer* Layer)
{
  FLayers.push_back(Layer);
  Layer->Canvas(this);
  if (FUpdateCount == 0)
    Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Reads the layout info stored in the given (XML) file and creates figure templates.
 * Existing templates remain in where they are but are replaced if a new definition with an existing name is found.
 *
 * @param Filename The name of the file to load.
 * @return Returns GC_NO_ERROR if everything was ok, otherwise an error code.
 */
TGCError CGenericCanvas::AddLayoutsFromFile(const char* Filename)
{
  TGCError Result = GC_NO_ERROR;
  xmlNodePtr Root, Current;
  CFigureParser Parser;

  string CurrentDir = GetCurrentDir();
  xmlDocPtr Document = xmlParseFile(Utf8ToANSI(string(Filename)).c_str());

  if (Document == NULL)
    return GC_XML_PARSE_ERROR;

  Root = xmlDocGetRootElement(Document);

  if (Root == NULL)
  {
    xmlFreeDoc(Document);
    Error("XML Error: Template file is empty.");
    return GC_XML_EMPTY_DOCUMENT;
  }
  
  if (!XML_IS(Root, "gc-layouts"))
  {
    xmlFreeDoc(Document);
    Error("XML Error: Template file invalid.");
    return GC_XML_INVALID_DOCUMENT;
  }

  // Switch to the directory of the given file. This is necessary to make relative file names working.
  string Path = ExtractFilePath(Filename);
  SetCurrentDir(Path);

  // Parse description elements.
  glMatrixMode(GL_MODELVIEW);
  Current = Root->children;
  while (Current != NULL)
  {
    // Be flexible, ignore any unknown layout entries.
    if (XML_IS(Current, "layout-definition"))
      Parser.ParseLayoutDefinition(Current, FModel);
    Current = Current->next;
  }

  SetCurrentDir(CurrentDir);
  xmlFreeDoc(Document);

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Adds a listener to the internal list.
 *
 * @param Listener The listener to add.
 */
void CGenericCanvas::AddListener(CGCListener* Listener)
{
  FListeners.push_back(Listener);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Type neutral helper method to add a new list item to any of the lists in the canvas. This method is meant to be used
 * by scripting engines and things like that because it is not as effective as direct manipulations.
 *
 * @param Path The path to a list property that gets a new entry.
 * @param Name The name of the new list entry. Other properties must be set via CGenericCanvas::Property.
 * @returns -1 if Path does not address a list property, otherwise the index of the newly added entry.
 */
int CGenericCanvas::AddNewItem(string Path, string Name)
{
  int Result = -1;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_VIEWS:
      {
        if (Path.empty())
        {
          Result = FViews.size();
          CreateView(Name.c_str()); // Also adds the new view to the views list.
          Change(this, GC_CHANGE_CANVAS_ADD_VIEW);
        }
        else
        {
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FViews.size()))
          {
            CGCView* View = FViews[Index];
            Result = View->AddNewItem(Path, Name);
          };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
        {
          Result = FLayers.size();
          CreateLayer(Name.c_str(), GC_LAYER_NORMAL, true); // Also adds the new layer to the layers list and the current view.
          Change(this, GC_CHANGE_CANVAS_ADD_LAYER);
        }
        else
        {
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            CLayer* Layer = FLayers[Index];
            Result = Layer->AddNewItem(Path, Name);
          };
        };
        break;
      };
    case GC_CONTAINER_MODEL:
      {
        Result = FModel->AddNewItem(Path, Name);
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Reads the style template info stored in the given (XML) file and creates style templates (OpenGL display lists).
 * Existing templates remain where they are but are replaced if a new definition with an existing name is found.
 *
 * @param Filename The name of the file to load.
 * @return Returns GC_NO_ERROR if everything was ok, otherwise an error code.
 */
TGCError CGenericCanvas::AddStylesFromFile(const char* Filename)
{
  TGCError Result = GC_NO_ERROR;
  xmlNodePtr Root, Current;
  CSVGParser Parser;

  string CurrentDir = GetCurrentDir();
  xmlDocPtr Document = xmlParseFile(Utf8ToANSI(string(Filename)).c_str());

  if (Document == NULL)
    return GC_XML_PARSE_ERROR;

  Root = xmlDocGetRootElement(Document);

  if (Root == NULL)
  {
    xmlFreeDoc(Document);
    Error("XML Error: Template file is empty.");
    return GC_XML_EMPTY_DOCUMENT;
  }
  
  if (!XML_IS(Root, "gc-styles"))
  {
    xmlFreeDoc(Document);
    Error("XML Error: Template file invalid.");
    return GC_XML_INVALID_DOCUMENT;
  }

  // Switch to the directory of the given file. This is necessary to make relative file names working.
  string Path = ExtractFilePath(Filename);
  SetCurrentDir(Path);

  // Parse description elements.
  glMatrixMode(GL_MODELVIEW);
  Current = Root->children;
  while (Current != NULL)
  {
    // Be flexible, ignore any unknown layout entries.
    if (XML_IS(Current, "style-definition"))
    {
      Parser.ParseDefinition(Current, FModel);
    }
    else
      if (XML_IS(Current, "texture"))
      {
        // Adds a new texture to the texture list.
        ParseTextureEntry(Current);
        CheckError();
      };
    Current = Current->next;
  }

  SetCurrentDir(CurrentDir);
  xmlFreeDoc(Document);

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds the given figure instance to the current selection. This call is simply forwared to the selection layer.
 * 
 * @param Instance The instance to be added to the selection. It is taken care that an instance is only added once.
 */
void CGenericCanvas::AddToSelection(CFigureInstance* Instance)
{
  FFeedbackLayer->AddToSelection(Instance);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Increases the update count by 1 to stop any recursive update until (@see EndUpdate()) was called.
 */
void CGenericCanvas::BeginUpdate(void)
{
  ++FUpdateCount;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Notifies all registered listeners that a change has occured.
 *
 * @param AObject The reference to an object for which the change event is triggered. Can be anything like a figure, figure
 *                instance, the canvas itself, a layer etc.
 * @param Reason Indicates what change actually happend.
 */
void CGenericCanvas::Change(void* AObject, TGCChangeReason Reason)
{
  CListenerIterator Iterator = FListeners.begin(); 
  while (Iterator != FListeners.end())
  {
    CListenerIterator Next = Iterator + 1;
    try
    {
      CGCListener* Listener = *Iterator;
      Listener->OnChange(AObject, Reason);
    }
    catch(...)
    {
      // If there was an exception while executing the method the listener is removed from our list
      // to avoid further harm.
      FListeners.erase(Iterator);
    };
    Iterator = Next;
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Checks if there is an OpenGL error registered and triggers the error method if so.
 */
void CGenericCanvas::CheckError(void)
{
  GLenum OGLError = glGetError();
  if (OGLError != GL_NO_ERROR)
  {
    char Buffer[1000];
    #ifdef _WINDOWS
      // Microsoft has confirmed that there is no guaranteed NULL termination (see MSDN for details).
      // Hence we better add one.
      _snprintf(Buffer, sizeof(Buffer), "OpenGL error encountered (0x%x).\0", (int) OGLError);
    #else
      snprintf(Buffer, sizeof(Buffer), "OpenGL error encountered (0x%x).", (int) OGLError);
    #endif // #ifdef _WINDOWS
    Error(Buffer);
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines the child count of the element given by Path.
 *
 * @param The path to the element whose child count is to be determined.
 * @return The number of childs or 0 if Path does not point to a list of child objects.
 */
int CGenericCanvas::ChildCount(string Path)
{
  int Result = 0;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_VIEWS:
      {
        if (Path.empty())
          Result = FViews.size();
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FViews.size()))
          {
            CGCView* View = FViews[Index];
            Result = View->ChildCount(Path);
          };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
          Result = FLayers.size();
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            CLayer* Layer = FLayers[Index];
            Result = Layer->ChildCount(Path);
          };
        };
        break;
      };
    case GC_CONTAINER_MODEL:
      {
        Result = FModel->ChildCount(Path);
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes all GC content.
 */
void CGenericCanvas::ClearContent(void)
{
  FModel->ClearFigures();
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes all layout info.
 */
void CGenericCanvas::ClearLayouts(void)
{
  FModel->ClearLayouts();
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes all currently selected figure instances from the selection set. This call is simply forwareded to the
 * selection layer.
 */
void CGenericCanvas::ClearSelection(void)
{
  FFeedbackLayer->ClearSelection();
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Causes the model to clear its style definitions.
 */
void CGenericCanvas::ClearStyles(void)
{
  FModel->ClearStyles();
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

CFigure* CGenericCanvas::CreateFigure(const char* Type)
{
  return FModel->CreateFigure(Utf8ToUtf16(Type));
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a new layer with the given name and returns it to the caller. The new layer is added to this canvas.
 *
 * @param Name The layer identification, encoded in UTF-8.
 * @param Type The type of layer that is requested.
 * @param AddToCurrentView If true then the new layer will be added to the current view (if there is any).
 * @return The new layer.
 */
CLayer* CGenericCanvas::CreateLayer(const char* Name, TGCLayerType Type, bool AddToCurrentView)
{
  CLayer* Layer = NULL;
  switch (Type) 
  {
    case GC_LAYER_NORMAL:
      {
        Layer = new CLayer(Name);
        break;
      };
    case GC_LAYER_GRID:
      {
        Layer = new CGridLayer(Name);
        break;
      };
  };

  if (Layer != NULL)
  {
    AddLayer(Layer);
    if (FCurrentView != NULL)
      FCurrentView->AddLayer(Layer);
  };

  return Layer;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a new view in this canvas, adds it to the internal list and returns it to the caller. If there is 
 * currently no current view set then the new view becomes the current one.
 *
 * @param Name The identification of the view (encoded in UTF-8). Should be unique.
 * @return The new view.
 */
CGCView* CGenericCanvas::CreateView(const char* Name)
{
  CGCView* View = new CGCView(this, Name);
  FViews.push_back(View);
  if (FCurrentView == NULL)
    CurrentView(View);

  return View;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the currently active view.
 *
 * @param View The new view to activate.
 */
void CGenericCanvas::CurrentView(CGCView* View)
{
  if (FCurrentView != View)
  {
    FCurrentView = View;
    if (FCurrentView != NULL)
      FStates |= GC_STATE_PENDING_ACTIVATION;
    else
      FStates &= ~GC_STATE_PENDING_ACTIVATION;
    Change(FCurrentView, GC_CHANGE_CANVAS_SWITCH_VIEW);
    Invalidate();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the currently active view.
 *
 * @return The currently active view.
 */
CGCView* CGenericCanvas::CurrentView(void)
{
  return FCurrentView;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * The counterpart to (@see BeginUpdate). It releases one update lock and invalidates the canvas if the count drops to 0.
 */
void CGenericCanvas::EndUpdate(void)
{
  if (FUpdateCount > 0)
    FUpdateCount--;
  if (FUpdateCount == 0)
    Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Notifies all registered listeners about an error that occured.
 */
void CGenericCanvas::Error(const char* Message)
{
  CListenerIterator Iterator = FListeners.begin(); 
  while (Iterator != FListeners.end())
  {
    CListenerIterator Next = Iterator + 1;
    try
    {
      CGCListener* Listener = *Iterator;
      Listener->OnError(Message);
    }
    catch(...)
    {
      // If there was an exception while executing the method the listener is removed from our list
      // to avoid further harm.
      FListeners.erase(Iterator);
    };
    Iterator = Next;
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates and returns a new figure instance enumerator instance.
 *
 * @return The new enumerator.
 */
CFigureInstanceEnumerator* CGenericCanvas::GetFigureInstanceEnumerator(void)
{
  return new CFigureInstanceEnumerator(this);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines whether the given position corresponds to any of the parts (body, handles) of a selection decoration.
 * This test is quite fast and can be used for cursor feedback and such. This method is simply forwarded to
 * CFeedbackLayer::GetSelectionInfo after modelview and projection matrix are constructed with the current settings.
 *
 * @param X The horizontal mouse position in window coordinates.
 * @param Y The vertical mouse position in window coordinate.
 * @return One of the direction flags.
 */
TGCDirection CGenericCanvas::GetSelectionInfo(const int X, const int Y)
{
  if (FCurrentView != NULL)
    FCurrentView->ApplyTransformations(true);

  return FFeedbackLayer->GetSelectionInfo(X, Y);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Notifies all registered listeners that a change has occured and they need to invalidate their viewers.
 */
void CGenericCanvas::Invalidate(void)
{
  CListenerIterator Iterator = FListeners.begin(); 
  while (Iterator != FListeners.end())
  {
    CListenerIterator Next = Iterator + 1;
    try
    {
      CGCListener* Listener = *Iterator;
      Listener->OnInvalidate();
    }
    catch(...)
    {
      // If there was an exception while executing the method the listener is removed from our list
      // to avoid further harm.
      FListeners.erase(Iterator);
    };
    Iterator = Next;
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Tells the selection layer to recompute the bounds of the selection decoration for the given figure instance.
 *
 * @param Instance The figure instance for which to recompute the selection decoration.
 */
void CGenericCanvas::InvalidateSelectionBounds(CFigureInstance* Instance)
{
  FFeedbackLayer->InvalidateBounds(Instance);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Tells the caller whether the canvas is currently being updated.
 */
bool CGenericCanvas::IsUpdating(void)
{
  return FUpdateCount != 0;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the first layer with the given name from the layers collection.
 *
 * @param Name The name of the layer to return.
 * @return The first found layer with the given name or NULL if no layer could be found.
 */
CLayer* CGenericCanvas::LayerByName(const char* Name)
{
  CLayer* Result = NULL;

  wstring S = Utf8ToUtf16(Name);
  for (CLayers::const_iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
  {
    CLayer* Layer = *Iterator;
    if (Layer->Name() == S)
    {
      Result = Layer;
      break;
    };
  };
  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Retrieves the value of the property given by path. The path syntax is must be something like (here expressed as regex)
 * (container)*(property), where container is a slash and the name of a container class (e.g. layers, figures) and
 * property is the name of a simple property of that container.
 *
 * @param Path The path of the property to return.
 * @return A description of the property value and, if the property is simple, the actual value.
 */
TGCVariant CGenericCanvas::Property(string Path)
{
  TGCVariant Result;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_ROOT:
      {
        switch (GetPropertyID(Path))
        {
          case GC_PROPERTY_NAME:
            {
              Result.Type = GC_VAR_STRING;
              Result.S = "Generic canvas";
              break;
            };
          case GC_PROPERTY_DESCRIPTION:
            {
              Result.Type = GC_VAR_STRING;
              Result.S = "A generic 2D canvas based on OpenGL.";
              break;
            };
        };
        break;
      };
    case GC_CONTAINER_VIEWS:
      {
        if (Path.empty())
          Result.Type = GC_VAR_LIST;
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FViews.size()))
          {
            if (Path.empty())
              // The entry object is addressed (since no property of it was specified).
              Result.Type = GC_VAR_OBJECT;
            else
            {
              CGCView* View = FViews[Index];
              Result = View->Property(Path);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
          Result.Type = GC_VAR_LIST;
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            if (Path.empty())
              // The entry object is addressed (since no property of it was specified).
              Result.Type = GC_VAR_OBJECT;
            else
            {
              CLayer* Layer = FLayers[Index];
              Result = Layer->Property(Path);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_MODEL:
      {
        if (Path.empty())
          Result.Type = GC_VAR_OBJECT;
        else
          Result = FModel->Property(Path);
        break;
      };
  };
  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Set the value of the given property, which must be a simple property.
 *
 * @param The path of the property.
 * @param Value The new value of the property. Automatic conversion is performed where possible.
 */
void CGenericCanvas::Property(string Path, const TGCVariant& Value)
{
  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_ROOT:
      {
        switch (GetPropertyID(Path))
        {
          // There are no simple properties in the canvas that can be changed.
          // GC_PROPERTY_NAME is only a placeholder to easy future enhancements.
          case GC_PROPERTY_NAME:
            {
              break;
            };
        };
        break;
      };
    case GC_CONTAINER_VIEWS:
      {
        if (!Path.empty())
        {
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FViews.size()))
          {
            if (!Path.empty())
            {
              CGCView* View = FViews[Index];
              View->Property(Path, Value);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (!Path.empty())
        {
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            if (!Path.empty())
            {
              CLayer* Layer = FLayers[Index];
              Layer->Property(Path, Value);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_MODEL:
      {
        if (!Path.empty())
          FModel->Property(Path, Value);
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Destroys this canvas instance. The release method is used by non-C++ languages access this code in order to avoid
 * releasing memory in an environment where it wasn't allocated.
 */
void CGenericCanvas::Release(void)
{
  delete this;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the given figure instance from the current selection. This call is simply forwared to the selection layer.
 *
 * @param Instance The figure instance to be removed from the selection. Nothing happens if it isn't selected.
 */
void CGenericCanvas::RemoveFromSelection(CFigureInstance* Instance)
{
  FFeedbackLayer->RemoveFromSelection(Instance);
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes an entry from one of our lists (given by path).
 *
 * @param Path The path to the list to modify.
 * @param Index The index of the entry to remove.
 */
void CGenericCanvas::RemoveItem(string Path, int Index)
{
  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_VIEWS:
      {
        if (Path.empty())
        {
          delete FViews[Index];
          FViews.erase(FViews.begin() + Index);
          Change(this, GC_CHANGE_CANVAS_REMOVE_VIEW);
        }
        else
        {
          int ViewIndex = GetEntryIndex(Path);
          if ((ViewIndex > -1) && (ViewIndex < (int) FViews.size()))
          {
            CGCView* View = FViews[ViewIndex];
            View->RemoveItem(Path, Index);
          };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
        {
          delete FLayers[Index];
          FLayers.erase(FLayers.begin() + Index);
          Change(this, GC_CHANGE_CANVAS_REMOVE_LAYER);
        }
        else
        {
          int LayerIndex = GetEntryIndex(Path);
          if ((LayerIndex > -1) && (LayerIndex < (int) FLayers.size()))
          {
            CLayer* Layer = FLayers[LayerIndex];
            Layer->RemoveItem(Path, Index);
          };
        };
        break;
      };
    case GC_CONTAINER_MODEL:
      {
        FModel->RemoveItem(Path, Index);
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the given layer from the internal layer list. The layer itself will not be destroed, just removed.
 *
 * @param Layer The layer to be removed.
 */
void CGenericCanvas::RemoveLayer(CLayer* Layer)
{
  for (CViews::const_iterator Iterator = FViews.begin(); Iterator != FViews.end(); ++Iterator)
    (*Iterator)->RemoveLayer(Layer);

  Layer->Canvas(NULL);
  for (CLayers::iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    if (*Iterator == Layer)
    {
      FLayers.erase(Iterator);
      break;
    };
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes a listener from the internal list. If the listener is not in the list then this call has no effect.
 *
 * @param Listener The listener to remove.
 */
void CGenericCanvas::RemoveListener(CGCListener* Listener)
{
  for (CListenerIterator Iterator = FListeners.begin(); Iterator != FListeners.end(); ++Iterator)
    if (*Iterator == Listener)
    {
      FListeners.erase(Iterator);
      break;
    };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the given view from the internal list.
 *
 * @param View The view to be removed.
 */
void CGenericCanvas::RemoveView(CGCView* View)
{
  if (FCurrentView == View)
    FCurrentView = NULL;
  for (CViews::iterator Iterator = FViews.begin(); Iterator != FViews.end(); ++Iterator)
    if (*Iterator == View)
    {
      FViews.erase(Iterator);
      break;
    };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * This is the main paint routine. It must be called by the viewer holding the reference to this canvas (e.g. when a 
 * window must be redrawn).
 */
void CGenericCanvas::Render(void)
{
  // No display if the canvas is currently being updated.
  if (FUpdateCount == 0)
  {
    if ((FStates & GC_STATE_PENDING_ACTIVATION) != 0)
    {
      // A pending activation state always means there is a valid current view.
      FStates &= ~GC_STATE_PENDING_ACTIVATION;
      FCurrentView->Activate();
    };

    ClearBuffers();
    if (FCurrentView != NULL)
      FCurrentView->Render();
    FFeedbackLayer->Render();
    CheckError();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Preparation function for CFeedbackLayer::ResizeFiguresStart.
 */
void CGenericCanvas::ResizeFiguresStart(int X, int Y, TGCDirection Direction)
{
  // Initialize our transformation queue for convertion of the given screen coordinates into layer
  // coordinates (which happens in the feedback layer).
  if (FCurrentView != NULL)
    FCurrentView->ApplyTransformations(true);

  FFeedbackLayer->ResizeFiguresStart(X, Y, Direction);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Preparation function for CFeedbackLayer::ResizeFiguresStop.
 */
void CGenericCanvas::ResizeFiguresStop(void)
{
  FFeedbackLayer->ResizeFiguresStop();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Preparation function for CFeedbackLayer::ResizeFiguresTo.
 */
void CGenericCanvas::ResizeFiguresTo(int X, int Y)
{
  if (FCurrentView != NULL)
    FCurrentView->ApplyTransformations(false);

  FFeedbackLayer->ResizeFiguresTo(X, Y);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Preparation function for CFeedbackLayer::RubberbandResize.
 */
void CGenericCanvas::RubberbandResize(int X, int Y, TRBSelectionAction Action)
{
  if (FCurrentView != NULL)
    FCurrentView->ApplyTransformations(false);

  FFeedbackLayer->RubberbandResize(X, Y, Action);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Preparation function for CFeedbackLayer::RubberbandStart.
 */
void CGenericCanvas::RubberbandStart(TRubberbandStyle Style, int X, int Y, bool ClearSelection)
{
  // Initialize our transformation queue for conversion of the given screen coordinates into layer
  // coordinates (which happens in the feedback layer).
  if (FCurrentView != NULL)
    FCurrentView->ApplyTransformations(true);

  FFeedbackLayer->RubberbandStart(Style, X, Y, ClearSelection);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Preparation function for CFeedbackLayer::RubberbandStop.
 */
void CGenericCanvas::RubberbandStop(void)
{
  FFeedbackLayer->RubberbandStop();
}

//----------------------------------------------------------------------------------------------------------------------

void CGenericCanvas::ShowSelection(bool Visible)
{
  FFeedbackLayer->Visible(Visible);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the first view with the given name from the views collection.
 *
 * @param Name The name of the view to return.
 * @return The first found view with the given name or NULL if no view could be found.
 */
CGCView* CGenericCanvas::ViewByName(const char* Name)
{
  CGCView* Result = NULL;

  for (CViews::const_iterator Iterator = FViews.begin(); Iterator != FViews.end(); ++Iterator)
  {
    CGCView* View = *Iterator;
    if (View->FName == Name)
    {
      Result = View;
      break;
    };
  };
  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

