/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_model.cpp 
 * @brief Implementation of the model that manages the visual representation in the generic canvas.
 * 
 */

#include "myx_gc.h"

#include "myx_gc_figure.h"
#include "myx_gc_model.h"
#include "myx_gc_canvas.h"

//----------------------------------------------------------------------------------------------------------------------

CGCModel::CGCModel(CGenericCanvas* Controller)
{
  FCanvas = Controller;
}

//----------------------------------------------------------------------------------------------------------------------

CGCModel::~CGCModel(void)
{
  ClearFigures();
  ClearLayouts();
  ClearStyles();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Checks if there is any pending OpenGL error (e.g. from style creation). Call is forwared to the canvas and handeled there.
 */
void CGCModel::CheckError(void)
{
  FCanvas->CheckError();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Type neutral helper method to add a new list item to any of the lists in the canvas. This method is meant to be used
 * by scripting engines and things like that because it is not as effective as direct manipulations.
 *
 * @param Path The path to a list property that gets a new entry.
 * @param Name The name or type of the new list entry. Other properties must be set via CGCModel::Property.
 * @returns -1 if Path does not address a list property, otherwise the index of the newly added entry.
 */
int CGCModel::AddNewItem(string Path, string Name)
{
  int Result = -1;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_FIGURES:
      {
        if (Path.empty())
        {
          Result = FFigures.size();
          CreateFigure(Utf8ToUtf16(Name));
          Change(this, GC_CHANGE_MODEL_ADD_FIGURE);
        };
        break;
      };
    case GC_CONTAINER_STYLES:
      {
        if (Path.empty())
        {
          wstring ID = Utf8ToUtf16(Name);
          CStyleList::iterator Iterator = FStyles.find(ID);
          if (Iterator == FStyles.end())
          {
            CGCStyle* NewStyle = new CGCStyle(ID);
            FStyles[ID] = NewStyle;

            // TODO: Avoid indexed access.
            Result = 0;
            for (Iterator = FStyles.begin(); Iterator != FStyles.end(); ++Iterator)
            {
              if (Iterator->first == ID)
                break;
              ++Result;
            };
            Change(this, GC_CHANGE_MODEL_ADD_STYLE);
          };
        };
        break;
      };
    case GC_CONTAINER_LAYOUTS:
      {
        // Currently layouts are only added while parsing the XML templates.
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Called when something changed in the model.
 *
 * @param AObject The reference to an object for which the change event is triggered (usually the model).
 * @param Reason Indicates what change actually happend.
 */
void CGCModel::Change(void* AObject, TGCChangeReason Reason)
{
  if (FCanvas != NULL)
    FCanvas->Change(AObject, Reason);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines the child count of the element given by Path.
 *
 * @param The path to the element whose child count is to be determined.
 * @return The number of childs or 0 if Path does not point to a list of child objects.
 */
int CGCModel::ChildCount(string Path)
{
  int Result = 0;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_FIGURES:
      {
        if (Path.empty())
          Result = FFigures.size();
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FFigures.size()))
          {
            CFigure* Figure = FFigures[Index];
            Result = Figure->ChildCount(Path);
          };
        };
        break;
      };
    case GC_CONTAINER_STYLES:
      {
        if (Path.empty())
          Result = FStyles.size();
        break;
      };
    case GC_CONTAINER_LAYOUTS:
      {
        if (Path.empty())
          Result = FLayouts.size();
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

CGenericCanvas* CGCModel::GetController(void)
{
  return FCanvas;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a new figure and puts in into the internal list.
 *
 * @param Type The name of the template from which to create the figure.
 */
CFigure* CGCModel::CreateFigure(wstring Type)
{
  CFigureTemplate* Template = Layout(Type);
  CFigure* Figure = new CFigure(this, Template);
  FFigures.push_back(Figure);

  return Figure;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Clears the model, that is, the figures defined in this model.
 */
void CGCModel::ClearFigures(void)
{
  FCanvas->BeginUpdate();
  for (CFigureList::iterator Iterator = FFigures.begin(); Iterator != FFigures.end(); ++Iterator)
    delete *Iterator;
  FFigures.clear();
  FCanvas->EndUpdate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Clears the figure template list (layouts).
 *
 * @note: Existing figures are not concerned by deleting the template list.
 */
void CGCModel::ClearLayouts(void)
{
  for (CLayoutList::iterator Iterator = FLayouts.begin(); Iterator != FLayouts.end(); ++Iterator)
    delete Iterator->second;
  FLayouts.clear();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Clears all defined styles.
 *
 * @note: Deleting all styles means to leave all figures without visual representation.
 */
void CGCModel::ClearStyles(void)
{
  for (CStyleList::iterator Iterator = FStyles.begin(); Iterator != FStyles.end(); ++Iterator)
    delete Iterator->second;
  FStyles.clear();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the layout entry for the given identifier. If ther eis no layout with that name yet then one is created.
 *
 * @param ID The identification of the style.
 * @param The corresponding layout entry.
 */
CFigureTemplate* CGCModel::Layout(wstring ID)
{
  CFigureTemplate* Result = NULL;

  CLayoutList::iterator Iterator = FLayouts.find(ID);
  if (Iterator == FLayouts.end())
  {
    Result = new CFigureTemplate(ID);
    FLayouts[ID] = Result;
  }
  else
    Result = Iterator->second;

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Retrieves the value of the property given by path. The path syntax is must be something like (here expressed as regex)
 * (container)*(property), where container is a slash and the name of a container class (e.g. layers, figures) and
 * property is the name of a simple property of that container.
 *
 * @param Path The path of the property to return.
 * @return A description of the property value and, if the property is simple, the actual value.
 */
TGCVariant CGCModel::Property(string Path)
{
  TGCVariant Result;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_FIGURES:
      {
        if (Path.empty())
          Result.Type = GC_VAR_LIST;
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FFigures.size()))
          {
            if (Path.empty())
              // The entry object is addressed (since no property of it was specified).
              Result.Type = GC_VAR_OBJECT;
            else
            {
              CFigure* Figure = FFigures[Index];
              Result = Figure->Property(Path);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_STYLES:
      {
        if (Path.empty())
          Result.Type = GC_VAR_LIST;
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if (Path.empty())
            Result.Type = GC_VAR_OBJECT;
          else
          {
            // TODO: Rewrite access using associative containers in GRT and key access here.
            if ((Index > -1) && (Index < (int) FStyles.size()))
            {
              for (CStyleList::const_iterator Iterator = FStyles.begin(); Iterator != FStyles.end() && Index > 0; ++Iterator, --Index)
                ;
              CGCStyle* Style = Iterator->second;
              Result = Style->Property(Path);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_LAYOUTS:
      {
        if (Path.empty())
          Result.Type = GC_VAR_LIST;
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);

          if (Path.empty())
            Result.Type = GC_VAR_OBJECT;
          else
          {
            // TODO: Rewrite access using associative containers in GRT and key access here.
            if ((Index > -1) && (Index < (int) FLayouts.size()))
            {
              for (CLayoutList::const_iterator Iterator = FLayouts.begin(); Iterator != FLayouts.end() && Index > 0; ++Iterator, --Index)
                ;
              CFigureTemplate* Template = Iterator->second;
              // Figure templates are only known by their name. So we handle this single case here.
              if (GetPropertyID(Path) == GC_PROPERTY_NAME)
              {
                Result.Type = GC_VAR_STRING;
                Result.S = Utf16ToUtf8(Template->ID());
              };
            };
          };
        };
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the value of the given property, which must be a simple property.
 *
 * @param The path of the property.
 * @param Value The new value of the property. Automatic conversion is performed where possible.
 */
void CGCModel::Property(string Path, const TGCVariant& Value)
{
  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_FIGURES:
      {
        if (!Path.empty())
        {
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FFigures.size()))
          {
            if (!Path.empty())
            {
              CFigure* Figure = FFigures[Index];
              Figure->Property(Path, Value);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_STYLES:
      {
        if (!Path.empty())
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if (!Path.empty())
          {
            // TODO: Rewrite access using associative containers in GRT and key access here.
            if ((Index > -1) && (Index < (int) FStyles.size()))
            {
              for (CStyleList::const_iterator Iterator = FStyles.begin(); Iterator != FStyles.end() && Index > 0; ++Iterator, --Index)
                ;
              CGCStyle* Style = Iterator->second;
              Style->Property(Path, Value);
            };
          };
        };
        break;
      };
    case GC_CONTAINER_LAYOUTS:
      {
        if (!Path.empty())
        {
          int Index = GetEntryIndex(Path);

          if (!Path.empty())
          {
            // TODO: Rewrite access using associative containers in GRT and key access here.
            if ((Index > -1) && (Index < (int) FLayouts.size()))
            {
              for (CLayoutList::const_iterator Iterator = FLayouts.begin(); Iterator != FLayouts.end() && Index > 0; ++Iterator, --Index)
                ;
              CFigureTemplate* Template = Iterator->second;
              
              // There is nothing that could be changed currently in a figure template.
            };
          };
        };
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes an entry from one of our lists (given by path).
 *
 * @param Path The path to the list to modify.
 * @param Index The index of the entry to remove.
 */
void CGCModel::RemoveItem(string Path, int Index)
{
  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_FIGURES:
      {
        if (Path.empty())
        {
          delete FFigures[Index];
          FFigures.erase(FFigures.begin() + Index);
          Change(this, GC_CHANGE_MODEL_REMOVE_FIGURE);
        };
        break;
      };
    case GC_CONTAINER_STYLES:
      {
        if (Path.empty())
        {
          // TODO: Avoid indexed access.
          CStyleList::iterator Iterator = FStyles.begin();
          for (; Index > 0, Iterator != FStyles.end(); --Index, ++Iterator)
            ;
          delete Iterator->second;
          FStyles.erase(Iterator);
          Change(this, GC_CHANGE_MODEL_ADD_STYLE);
        };
        break;
      };
    case GC_CONTAINER_LAYOUTS:
      {
        // Currently layouts can only be removed as a whole by calling CGCModel::ClearLayouts.
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the style entry for the given identifier. If there is no style with that name one is created.
 *
 * @param ID The identification of the style.
 * @return The corresponding style entry, can be NULL if ID is empty.
 */
CGCStyle* CGCModel::Style(wstring ID)
{
  CGCStyle* Result = NULL;

  if (ID.size() > 0)
  {
    CStyleList::iterator Iterator = FStyles.find(ID);
    if (Iterator == FStyles.end())
    {
      Result = new CGCStyle(ID);
      FStyles[ID] = Result;
    }
    else
      Result = Iterator->second;
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

