/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_utilities.cpp
 * @brief Some common utility functions.
 * 
 */

#include "myx_gc.h"

#include <errno.h>
#include <glib.h>

#include "myx_gc_utilities.h"
#include "myx_gc_const.h"

//----------------------------------------------------------------------------------------------------------------------

/**
 * Converts the given string, which is supposed to be an UTF-8 encoded text into an UTF-16 string.
 *
 * @param Source Contains the source string encoded in UTF-8.
 * @return The converted string in UTF-16 encoding.
 */
wstring Utf8ToUtf16(const string& Source)
{
  wstring Result;

#ifdef _WINDOWS
  const char* Raw = Source.c_str();
  int Size = MultiByteToWideChar(CP_UTF8, 0, Raw, -1, NULL, 0);
  wchar_t* Buffer = new wchar_t[Size];
  MultiByteToWideChar(CP_UTF8, 0, Raw, -1, Buffer, Size);
  Result += Buffer;
  delete[] Buffer;
#else
  gunichar2 *tmp;
  tmp= g_utf8_to_utf16(Source.c_str(), Source.length(), NULL, NULL, NULL);
  Result= (wchar_t*)tmp;
  g_free(tmp);
#endif

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Converts the given UTF-16 string into an UTF-8 string.
 *
 * @param Source Contains the source string encoded in UTF-16.
 * @return The converted string in UTF-8 encoding.
 */
string Utf16ToUtf8(const wstring& Source)
{
  string Result;

#ifdef _WINDOWS
  const wchar_t* Raw = Source.c_str();
  int Size = WideCharToMultiByte(CP_UTF8, 0, Raw, -1, NULL, 0, NULL, NULL);
  char* Buffer = new char[Size];
  WideCharToMultiByte(CP_UTF8, 0, Raw, -1, Buffer, Size, NULL, NULL);
  Result += Buffer;
  delete[] Buffer;
#else
  gunichar2 *tmp;
  tmp= g_utf16_to_utf8((gunichar2*)Source.c_str(), Source.length(), NULL, NULL, NULL);
  Result= (char*)tmp;
  g_free(tmp);
#endif

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Converts the given string into an ANSI string.
 * @param Source Contains the source string encoded in UTF-16.
 * @return The converted string in ANSI encoding.
 * @note The current user locale is used to convert the Unicode string to ANSI.
 */
string Utf16ToANSI(const wstring& Source)
{
  string Result;

#ifdef _WINDOWS
  const wchar_t* Raw = Source.c_str();
  int Size = WideCharToMultiByte(CP_ACP, 0, Raw, -1, NULL, 0, NULL, NULL);
  char* Buffer = new char[Size];
  WideCharToMultiByte(CP_ACP, 0, Raw, -1, Buffer, Size, NULL, NULL);
  Result += Buffer;
  delete[] Buffer;
#else
  gchar *tmp;
  tmp= g_utf16_to_utf8((gunichar2*)Source.c_str(), Source.length(), NULL, NULL, NULL);
  Result= tmp;
  g_free(tmp);
#endif

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Converts the given string, which is supposed to be an UTF-8 encoded text into an ANSI string.
 * @param Source Contains the source string encoded in UTF-8.
 * @return The converted string in ANSI encoding.
 * @note The current user locale is used to convert the Unicode string to ANSI.
 */
string Utf8ToANSI(const string& Source)
{
#ifdef _WINDOWS
  return Utf16ToANSI(Utf8ToUtf16(Source));
#else
  return Source;
#endif
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief User defined file handling callback for handling file i/o in png lib.
 *        We have to make it so because of the FILE allocation and png being in a DLL on Windows.
 */
void PNGRead(png_structp png_ptr, png_bytep data, png_size_t length)
{
  fread(data, length, 1, (FILE*) png_ptr->io_ptr);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Loads a png file.
 * @param Filename An ANSI encoded file name (can contain path info) to the png file.
 * @return Returns a pointer to a TImage structure containing the image data.
 * @note The return memory must be freed using FreeImage().
 */
TImage* LoadPNG(const string& Filename)
{
  TImage *Image = NULL;
  FILE* File;
  png_structp PNG;
  png_infop Info;
  png_uint_32 Width, Height;
  int depth, junk, color_type;
  int Offset;

  File = fopen(Filename.c_str(), "rb");
  try
  {
    if (File) 
    {
      png_byte Magic[8];
      if (fread(Magic, sizeof(Magic), 1, File) != 1)
      {
        fprintf(stderr, "can't read '%s': %s\n", Filename.c_str(), strerror(errno));
        fclose(File);
        return NULL;
      };
      if (!png_check_sig(Magic, sizeof(Magic)))
      {
        fclose(File);
        return NULL;
      };

      PNG = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
      if (!PNG)
      {
        fclose(File);
        return NULL;
      };
      
      Info = png_create_info_struct(PNG);
      if (!Info) 
      {
        fclose(File);
        png_destroy_read_struct(&PNG, NULL, NULL);
        return NULL;
      };

      png_set_read_fn(PNG, File, &PNGRead);

      png_init_io(PNG, File);
      png_set_sig_bytes(PNG, sizeof(Magic));

      png_read_info(PNG, Info);

      png_get_IHDR(PNG, Info, &Width, &Height, &depth, &color_type,
                  &junk, &junk, &junk);

      /* sanity check */
      if (Width < 1 || Height < 1) 
      {
        fclose(File);
        png_destroy_read_struct(&PNG, &Info, NULL);
        return NULL;
      }

      Image = (TImage*) malloc(sizeof(TImage));
      if (!Image)
      {
        fclose(File);
        png_destroy_read_struct(&PNG, &Info, NULL);
        return NULL;
      };

      Image->Data = (unsigned char*) malloc(Width * Height * Info->channels);
      if (!Image->Data) 
      {
        fclose(File);
        png_destroy_read_struct(&PNG, &Info, NULL);
        return NULL;
      };

      Image->Width = Width;
      Image->Height = Height;
      Image->ColorType = (TColorType) Info->color_type;
      Image->Channels = Info->channels;

      // normalize to 8bpp with alpha channel
      if (color_type == PNG_COLOR_TYPE_PALETTE && depth <= 8)
        png_set_expand(PNG);

      if (color_type == PNG_COLOR_TYPE_GRAY && depth <= 8)
        png_set_expand(PNG);

      if (png_get_valid(PNG, Info, PNG_INFO_tRNS))
        png_set_expand(PNG);

      if (depth == 16)
        png_set_strip_16(PNG);

      if (color_type == PNG_COLOR_TYPE_GRAY ||
          color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
        png_set_gray_to_rgb(PNG);

      /* do the transforms */
      png_read_update_info(PNG, Info);

      Offset= 0;
      /* read Data */
      for (unsigned int I = 0; I < Height; I++)
      {
        png_read_row(PNG, &Image->Data[Offset], NULL);
        Offset += Info->channels * Width;
      }
      png_read_end(PNG, Info);
      png_destroy_read_struct(&PNG, &Info, NULL);

      fclose(File);
    };
  }
  catch (...)
  {
    if (Image)
      FreeImage(Image);
    Image = NULL;
  };

  return Image;
}

//----------------------------------------------------------------------------------------------------------------------

void FreeImage(TImage* Image)
{
  free(Image->Data);
  free(Image);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Returns the current working folder (ANSI encoded).
 * @return The current working folder.
 */
string GetCurrentDir(void)
{
  char Buffer[MAX_PATH];

  getcwd(Buffer, sizeof(Buffer));

  return string(Buffer);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Sets the current working folder (folder name must be ANSI encoded).
 * @param Folder The new folder to be set.
 */
void SetCurrentDir(const string& Folder)
{
  chdir(Folder.c_str());
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief ExtractFilePath extracts the drive and directory parts of the given filename. The resulting string is the 
          leftmost characters of FileName, up to and including the colon or backslash that separates the path information 
          from the name and extension. The resulting string is empty if FileName contains no drive and directory parts.
 * @param Filename The file name (ANSI encoded) of which the path is to be extracted.
 * @return The extracted path part (ANSI encoded).
 */
string ExtractFilePath(const string& Filename)
{
  gchar* Raw = g_path_get_dirname(Filename.c_str());

  string Result(Raw);
  g_free(Raw);

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Platform neutral file open routine.
 * @param Filename The name of file encode in UTF-8.
 * @param OpenMode The mode how to open the file (the same as used for fopen calls).
 * @return A pointer to a FILE structure if the file could be opened, NULL otherwise.
 */
FILE* OpenFile(string Filename, const char* OpenMode)
{
#ifdef _WINDOWS
  wstring Name = Utf8ToUtf16(Filename);
  wstring Mode = Utf8ToUtf16(string(OpenMode));
  return _wfopen(Name.c_str(), Mode.c_str());
#else
  FILE *file;
  char * local_filename;

  if (!(local_filename = g_filename_from_utf8(Filename.c_str(), -1, NULL, NULL, NULL)))
    return NULL;

  file = fopen(local_filename, OpenMode);

  g_free(local_filename);

  return file;
#endif // #ifdef _WINDOWS
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Helper method to sort left/right and bottom/top coordinates so that for left/top are always smaller than right/bottom
 * (origin is considered in the left-upper corner, +y pointing down).
 */
TBounds SortBounds(const TBounds& Bounds)
{
  TBounds Result;

  Result.Left = (Bounds.Left <= Bounds.Right) ? Bounds.Left : Bounds.Right;
  Result.Right = (Bounds.Left <= Bounds.Right) ? Bounds.Right : Bounds.Left;
  Result.Top = (Bounds.Top <= Bounds.Bottom) ? Bounds.Top : Bounds.Bottom;
  Result.Bottom = (Bounds.Top <= Bounds.Bottom) ? Bounds.Bottom : Bounds.Top;

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Determines whether both bounds overlap. 
 *
 * @param Bounds1 One of the bounds to compare.
 * @param Bounds2 The other bounds to compare.
 * @return True if both bounds overlap each other, otherwise false.
 */
bool BoundsIntersect(const TBounds& Bounds1, const TBounds& Bounds2)
{
  return (Bounds1.Right > Bounds2.Left) && (Bounds1.Left < Bounds2.Right) && 
    (Bounds1.Bottom > Bounds2.Top) && (Bounds1.Top < Bounds2.Bottom);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Examines the given bounds and returns whether it is empty or not.
 *
 * @param Bounds The bounds to examine.
 * @return True if the bounds are empty, false otherwise.
 */
bool BoundsAreEmpty(const TBounds& Bounds)
{
  return (Bounds.Left == Bounds.Right) && (Bounds.Top == Bounds.Bottom);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Extracts the left most path part (separated by a slash character) and returns an ID for it. This path part is then
 * also removed from the given path.
 *
 * @param Path [in, out] The path to examine. It must start with a slash to be valid. On return it contains a new 
 *                       path where the top level part (if there was one) is removed.
 * @return An identifier that specifies which container was given in the top path part.
 */
TContainerID GetTopContainer(string& Path)
{
  TContainerID Result = GC_CONTAINER_UNKNOWN;

#ifdef __GNUC__
  static map<string, TContainerID> Containers;
#else
  static hash_map<string, TContainerID> Containers;
#endif

  // Initialize container mapping if not yet done.
  if (Containers.size() == 0)
  {
    Containers["/"] = GC_CONTAINER_ROOT;
    Containers["/layers"] = GC_CONTAINER_LAYERS;
    Containers["/feedback"] = GC_CONTAINER_FEEDBACK;
    Containers["/views"] = GC_CONTAINER_VIEWS;
    Containers["/model"] = GC_CONTAINER_MODEL;
    Containers["/style"] = GC_CONTAINER_STYLE;
    Containers["/styles"] = GC_CONTAINER_STYLES;
    Containers["/layouts"] = GC_CONTAINER_LAYOUTS;
    Containers["/figure"] = GC_CONTAINER_FIGURE;
    Containers["/figures"] = GC_CONTAINER_FIGURES;
    Containers["/figureInstances"] = GC_CONTAINER_FIGURE_INSTANCES;
    Containers["/content"] = GC_CONTAINER_FIGURE_CONTENT;
    Containers["/scaling"] = GC_CONTAINER_SCALING;
    Containers["/translation"] = GC_CONTAINER_TRANSLATION;
    Containers["/rotation"] = GC_CONTAINER_ROTATION;
    Containers["/groups"] = GC_CONTAINER_GROUPS;
    Containers["/constraints"] = GC_CONTAINER_CONSTRAINTS;
    Containers["/children"] = GC_CONTAINER_CHILDREN;
    Containers["/caption"] = GC_CONTAINER_CAPTION;
  };

  if (Path.size() > 0 && Path[0] == '/')
  {
    // Is there another path separator after the initial one?
    string::size_type Position = Path.find_first_of('/', 1);
    string ContainerPath;
    if (Position != Path.npos)
    {
      ContainerPath = Path.substr(0, Position);
      Path = Path.substr(Position);
    }
    else
      ContainerPath = Path;
    if (Containers.find(ContainerPath) != Containers.end())
    {
      Result = Containers[ContainerPath];
      if (ContainerPath == Path)
        Path = "";
    }
    else
      Result = GC_CONTAINER_ROOT;
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Treats the given path as property name preceeded with a slash. This name is looked up in a map and an identifier
 * is returned that can be used in switch statement.
 *
 * @param Path The path to examine. It must start with a slash to be valid.
 * @return An identifier that specifies which property was given in the top path part.
 */
TPropertyID GetPropertyID(const string& Path)
{
#ifdef __GNUC__
  static map<string, TPropertyID> Properties;
  map<string, TPropertyID>::const_iterator Iterator;
#else
  static hash_map<string, TPropertyID> Properties;
  hash_map<string, TPropertyID>::const_iterator Iterator;
#endif
  TPropertyID Result = GC_PROPERTY_UNKNOWN;

  // Fill property mapping first if not yet done.
  if (Properties.size() == 0)
  {
    // Simple properties.
    Properties["/width"] = GC_PROPERTY_WIDTH;
    Properties["/height"] = GC_PROPERTY_HEIGHT;
    Properties["/x"] = GC_PROPERTY_X;
    Properties["/y"] = GC_PROPERTY_Y;
    Properties["/z"] = GC_PROPERTY_Z;
    Properties["/name"] = GC_PROPERTY_NAME;
    Properties["/description"] = GC_PROPERTY_DESCRIPTION;
    Properties["/zoomx"] = GC_PROPERTY_ZOOMX;
    Properties["/zoomy"] = GC_PROPERTY_ZOOMY;
    Properties["/color"] = GC_PROPERTY_COLOR;
    Properties["/jitter"] = GC_PROPERTY_JITTER;
    Properties["/angle"] = GC_PROPERTY_ANGLE;
    Properties["/visible"] = GC_PROPERTY_VISIBLE;
    Properties["/enabled"] = GC_PROPERTY_ENABLED;
    Properties["/selected"] = GC_PROPERTY_SELECTED;
    Properties["/layout"] = GC_PROPERTY_LAYOUT;
    Properties["/resizable"] = GC_PROPERTY_RESIZABLE;
    Properties["/expanded"] = GC_PROPERTY_EXPANDED;
    Properties["/minWidth"] = GC_PROPERTY_MIN_WIDTH;
    Properties["/minHeight"] = GC_PROPERTY_MIN_HEIGHT;
    Properties["/maxWidth"] = GC_PROPERTY_MAX_WIDTH;
    Properties["/maxHeight"] = GC_PROPERTY_MAX_HEIGHT;
    Properties["/text"] = GC_PROPERTY_TEXT;
    Properties["/fontFamily"] = GC_PROPERTY_FONT_FAMILY;
    Properties["/fontSize"] = GC_PROPERTY_FONT_SIZE;
    Properties["/fontWeight"] = GC_PROPERTY_FONT_WEIGHT;
    Properties["/fontStyle"] = GC_PROPERTY_FONT_STYLE;
    Properties["/verticalAlignment"] = GC_PROPERTY_ALIGNMENT_VERTICAL;
    Properties["/horizontalAlignment"] = GC_PROPERTY_ALIGNMENT_HORIZONTAL;
    Properties["/bidiMode"] = GC_PROPERTY_BIDI_MODE;
  };

  if (Properties.find(Path) != Properties.end())
    Result = Properties[Path];

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Treats the given path as property name preceeded with a slash. The first (or only) subpath must be an integer number
 * denoting an index in a list.
 *
 * @param Path [in, out] The path to examine. It must start with a slash to be valid. On return it contains a new 
 *                       path without the index part (can be empty then).
 * @return The integer value extracted from the top path part.
 */
int GetEntryIndex(string& Path)
{
  int Result = -1;

  if (Path.size() > 0 && Path[0] == '/')
  {
    // Is there another path separator after the initial one?
    string::size_type Position = Path.find_first_of('/', 1);
    string Index;
    if (Position != Path.npos)
    {
      Index = Path.substr(1, Position - 1);
      Path = Path.substr(Position);
    }
    else
    {
      Index = Path.substr(1);
      Path = "";
    };
    Result = atoi(Index.c_str());
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

static TColorMap PredefinedColors;

/**
 * Converts a color to a string in the form #RRGGBB.
 *
 * @param Color The color to convert.
 * @return The given color as HTML string.
 */
string ColorToString(GLfloat* Color)
{
  char Buffer[10];

  #ifdef _WINDOWS
    // Microsoft has confirmed that there is no guaranteed NULL termination (see MSDN for details).
    // Hence we better add one.
    _snprintf(Buffer, sizeof(Buffer), "#%.2X%.2X%.2X\0", ROUND(Color[0] * 255), ROUND(Color[1] * 255), ROUND(Color[2] * 255));
  #else
    snprintf(Buffer, sizeof(Buffer), "#%.2X%.2X%.2X", ROUND(Color[0] * 255), ROUND(Color[1] * 255), ROUND(Color[2] * 255));
  #endif // #ifdef _WINDOWS

  return Buffer;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Converts a color to a string in the form #RRGGBB.
 *
 * @param Color The color to convert.
 * @return The given color as HTML string.
 */
string ColorToString(GLubyte* Color)
{
  char Buffer[10];

  #ifdef _WINDOWS
    // Microsoft has confirmed that there is no guaranteed NULL termination (see MSDN for details).
    // Hence we better add one.
    _snprintf(Buffer, sizeof(Buffer), "#%.2X%.2X%.2X\0", Color[0], Color[1], Color[2]);
  #else
    snprintf(Buffer, sizeof(Buffer), "#%.2X%.2X%.2X", Color[0], Color[1], Color[2]);
  #endif // #ifdef _WINDOWS

  return Buffer;
}

//----------------------------------------------------------------------------------------------------------------------

// Converts the two hex digits given by Upper and Lower to an unsigned byte.
unsigned char HexToByte(char Upper, char Lower)
{
  Upper -= '0';
  if (Upper > 9)
    Upper -= 7;
  Lower -= '0';
  if (Lower > 9)
    Lower -= 7;
  return Upper * 16 + Lower;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Searchs the predifined colors and tries to find one with the given name.
 *
 * @param Name The name of the color to find.
 * @param Color [out] The color data if it could be found.
 * @returns true if the color could be found, otherwise false;
 */
bool ColorByName(string Name, GLubyte* Color)
{

  // Fill the color map if not already done.
  if (PredefinedColors.size() == 0)
  {
    for (unsigned int I = 0; I < COLOR_COUNT; I++)
      PredefinedColors[Colors[I].Name] = Colors[I].Color;

#ifdef _WINDOWS
    for (unsigned int I = 0; I < SYS_COLOR_COUNT; I++)
    {
      COLORREF Reference = GetSysColor(SystemColors[I].Value);
      SystemColors[I].Color[0] = GetRValue(Reference);
      SystemColors[I].Color[1] = GetGValue(Reference);
      SystemColors[I].Color[2] = GetBValue(Reference);
      SystemColors[I].Color[3] = 1;
      PredefinedColors[SystemColors[I].Name] = SystemColors[I].Color;
    };
#endif // #ifdef _WINDOWS
  };

  TColorMapIterator Iterator = PredefinedColors.find(Name);
  bool Result = Iterator != PredefinedColors.end();
  if (Result)
  {
    Color[0] = Iterator->second[0];
    Color[1] = Iterator->second[1];
    Color[2] = Iterator->second[2];
    Color[3] = Iterator->second[3];
  };

  return Result;
}


//----------------------------------------------------------------------------------------------------------------------

/**
 * Registers predifined colors.
 *
 * @param ColorMap Colors to add to the predefined color list.
 */
void RegisterSystemColors(const TColorMap &ColorMap)
{
  for (TColorMapIterator Iterator = ColorMap.begin(); Iterator != ColorMap.end(); ++Iterator)
    PredefinedColors[Iterator->first]= Iterator->second;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Converts a string to a color with byte members.
 * The allowed syntax for colors is (as given by the SVG specification) either an HTML like
 * value (e.g. #FFFFFF, #FFF) or a function like form (e.g. rgb(100, 255, 255), rgb(10%, 100%, 0%)).
 *
 * @param ColorString The string to parse.
 * @param [out] The color converted from the string. It must have room for at least 3 members.
 * @return 
 *   0 - If a color could be found and converted.
 *   1 - If a color could be found but a conversion error occured.
 *   2 - No color was given.
 *   3 - The special color "none" was found.
 */
int StringToColor(string ColorString, GLubyte* Color)
{
  int Result = 2;
  
  if (ColorString.size() > 0)
  {
    Color[0] = 0;
    Color[1] = 0; 
    Color[2] = 0; 
    const char* Head = ColorString.c_str();
    const char* Tail;

    // Start by skipping leading white spaces. We have only simple ASCII compatible strings here,
    // so we don't need to care for conversion from UTF-8.
    while ((*Head != '\0') && ((*Head == ' ') || (*Head == '\t')))
      Head++;

    if (*Head != '\0')
    {
      if (*Head == '#')
      {
        // HTML color.
        Head++;
        Tail = Head;
        while ((*Tail >= '0' && *Tail <= '9') || (*Tail >= 'a' && *Tail <= 'f') || (*Tail >= 'A' && *Tail <= 'F'))
          Tail++;
        switch (Tail - Head)
        {
          // Support only the two defined cases. Other values are simply ignored.
          case 3:
            {
              Color[0] = HexToByte(*Head, *Head);
              Head++;
              Color[1] = HexToByte(*Head, *Head);
              Head++;
              Color[2] = HexToByte(*Head, *Head);
              Result = 0;

              break;
            };
          case 6:
            {
              Tail = Head + 1;
              Color[0] = HexToByte(*Head, *Tail);
              Head += 2; Tail += 2;
              Color[1] = HexToByte(*Head, *Tail);
              Head += 2; Tail += 2;
              Color[2] = HexToByte(*Head, *Tail);
              Result = 0;

              break;
            };
        }
      }
      else
        if (strcmp(Head, "none") == 0)
        {
          // Do not fill at all.
          Result = 3;
        }
        else
        {
          bool isRGB = strncmp(Head, "rgb(", 4) == 0;
          if (isRGB)
          {
            // Found a function call like specification. Split the entries and look if they are absolute
            // or percentage values.
            int Index = 0;
            int Value;
            int ComponentCount = 3;
            Head += ComponentCount + 1;
            while (Index < ComponentCount)
            {
              Value = 0;
              
              // Skip leading white spaces.
              while ((*Head != '\0') && ((*Head == ' ') || (*Head == '\t')))
                Head++;
              while ((*Head >= '0') && (*Head <= '9'))
                Value = Value * 10 + (*Head++ - '0');

              if (Value < 0)
                Value = 0;
              if (*Head == '%')
              {
                if (Value > 100)
                  Value = 100;
                Value = (Value * 255) / 100;
                Head++;
              }
              else
              {
                if (Value > 255)
                  Value = 255;
              };

              Color[Index++] = Value;

              // Skip trailing white spaces.
              while ((*Head != '\0') && ((*Head == ' ') || (*Head == '\t')))
                Head++;

              // If there is no comma or closing parenthesis then there is something wrong.
              if (*Head == ')') 
                break;
              if (*Head != ',')
                return 1;

              Head++;
            };
            if ((*Head == ')') && (Index == ComponentCount))
              Result = 0;
            else
              Result = 1;
          }
          else
          {
            // Last chance are color names. Try to find the text in the color constants.
            if (ColorByName(Head, Color))
              Result = 0;
            else
              Result = 1;
          };
        };
    };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Converts a string to color with float members.
 * The allowed syntax for colors is (as given by the SVG specification) either an HTML like
 * value (e.g. #FFFFFF, #FFF) or a function like form (e.g. rgb(100, 255, 255), rgb(10%, 100%, 0%)).
 *
 * @param ColorString The string to parse.
 * @param [out] The color converted from the string. It must have room for at least 3 members.
 * @return 
 *   0 - If a color could be found and converted.
 *   1 - If a color could be found but a conversion error occured.
 *   2 - No color was given.
 *   3 - The special color "none" was found.
 */
int StringToColor(string ColorString, GLfloat* Color)
{
  GLubyte ByteColor[3];
  int Result = StringToColor(ColorString, ByteColor);
  if (Result == 0)
  {
    Color[0] = (float) ByteColor[0] / 255;
    Color[1] = (float) ByteColor[1] / 255;
    Color[2] = (float) ByteColor[2] / 255;
  };
  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Converts a GC variant to a string.
 *
 * @param Variant The variant to convert.
 * @return The value of the variant as string.
 */
string VariantToString(const TGCVariant& Variant)
{
  string Result;

  switch (Variant.Type)
  {
    case GC_VAR_UNKNOWN:
      {
        Result = "unknown";
        break;
      };
    case GC_VAR_BOOL:
      {
        Result = Variant.B ? "true" : "false";
        break;
      };
    case GC_VAR_INT:
      {
        char Buffer[12];
        Result = itoa(Variant.I, Buffer, 10);
        break;
      };
    case GC_VAR_FLOAT:
      {
        char Buffer[100];
        #ifdef _WINDOWS
          // Microsoft has confirmed that there is no guaranteed NULL termination (see MSDN for details).
          // Hence we better add one.
          _snprintf(Buffer, sizeof(Buffer), "%gi\0", Variant.F);
        #else
          snprintf(Buffer, sizeof(Buffer), "%g", Variant.F);
        #endif // #ifdef _WINDOWS

        Result = Buffer;
        break;
      };
    case GC_VAR_STRING:
      {
        Result = Variant.S;
        break;
      };
    case GC_VAR_LIST:
      {
        Result = "list";
        break;
      };
    case GC_VAR_OBJECT:
      {
        Result = "object";
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Converts a GC variant to an integer.
 *
 * @param Variant The variant to convert.
 * @return The value of the variant as integer.
 */
int VariantToInt(const TGCVariant& Variant)
{
  int Result = 0;

  switch (Variant.Type)
  {
    case GC_VAR_UNKNOWN:
      {
        Result = 0;
        break;
      };
    case GC_VAR_BOOL:
      {
        Result = Variant.B ? 1 : 0;
        break;
      };
    case GC_VAR_INT:
      {
        Result = Variant.I;
        break;
      };
    case GC_VAR_FLOAT:
      {
        Result = ROUND(Variant.F);
        break;
      };
    case GC_VAR_STRING:
      {
        Result = atoi(Variant.S.c_str());
        break;
      };
    case GC_VAR_LIST:
    case GC_VAR_OBJECT:
      {
        Result = 0;
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Converts a GC variant to a float.
 *
 * @param Variant The variant to convert.
 * @return The value of the variant as float.
 */
float VariantToFloat(const TGCVariant& Variant)
{
  float Result = 0;

  switch (Variant.Type)
  {
    case GC_VAR_UNKNOWN:
      {
        Result = 0;
        break;
      };
    case GC_VAR_BOOL:
      {
        Result = (float) (Variant.B ? 1 : 0);
        break;
      };
    case GC_VAR_INT:
      {
        Result = (float) Variant.I;
        break;
      };
    case GC_VAR_FLOAT:
      {
        Result = Variant.F;
        break;
      };
    case GC_VAR_STRING:
      {
        Result = (float) atof(Variant.S.c_str());
        break;
      };
    case GC_VAR_LIST:
    case GC_VAR_OBJECT:
      {
        Result = 0;
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Converts a GC variant to a bool.
 *
 * @param Variant The variant to convert.
 * @return The value of the variant as bool.
 */
bool VariantToBool(const TGCVariant& Variant)
{
  bool Result = false;

  switch (Variant.Type)
  {
    case GC_VAR_UNKNOWN:
      {
        Result = false;
        break;
      };
    case GC_VAR_BOOL:
      {
        Result = Variant.B;
        break;
      };
    case GC_VAR_INT:
      {
        Result = Variant.I > 0;
        break;
      };
    case GC_VAR_FLOAT:
      {
        Result = Variant.F > 0;
        break;
      };
    case GC_VAR_STRING:
      {
        if (Variant.S == "true" || Variant.S == "yes" || Variant.S == "t" || Variant.S == "y")
          Result = true;
        else
        {
          float Value = (float) atof(Variant.S.c_str());
          Result = Value > 0;
        };
        break;
      };
    case GC_VAR_LIST:
      {
        Result = false;
        break;
      };
    case GC_VAR_OBJECT:
      {
        Result = false;
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a GC variant from the given value.
 * 
 * @param Value The value to wrap in the variant.
 * @return A new variant carrying the given value.
 */
TGCVariant Variant(const string& Value)
{
  TGCVariant Result;

  Result.Type = GC_VAR_STRING;
  Result.S = Value;

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a GC variant from the given value.
 * 
 * @param Value The value to wrap in the variant.
 * @return A new variant carrying the given value.
 */
TGCVariant Variant(const int Value)
{
  TGCVariant Result;

  Result.Type = GC_VAR_INT;
  Result.I = Value;

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a GC variant from the given value.
 * 
 * @param Value The value to wrap in the variant.
 * @return A new variant carrying the given value.
 */
TGCVariant Variant(const float Value)
{
  TGCVariant Result;

  Result.Type = GC_VAR_FLOAT;
  Result.F = Value;

return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a GC variant from the given value.
 * 
 * @param Value The value to wrap in the variant.
 * @return A new variant carrying the given value.
 */
TGCVariant Variant(const bool Value)
{
  TGCVariant Result;

  Result.Type = GC_VAR_BOOL;
  Result.B = Value;

  return Result;
}

//----------------- CBoundingBoxComputer -------------------------------------------------------------------------------

CBoundingBoxComputer::CBoundingBoxComputer(void)
{
  // The bound box initializes itself to all zero.
}

//----------------------------------------------------------------------------------------------------------------------

CBoundingBoxComputer::CBoundingBoxComputer(const TBoundingBox& InitialBox)
{
  FCurrentBox = InitialBox;
}

//----------------------------------------------------------------------------------------------------------------------

void CBoundingBoxComputer::IncludeVertex(const TVertex& Vertex)
{
  if (Vertex.X < FCurrentBox.UpperLeftX)
    FCurrentBox.UpperLeftX = Vertex.X;
  if (Vertex.X > FCurrentBox.LowerRightX)
    FCurrentBox.LowerRightX = Vertex.X;
  if (Vertex.Y < FCurrentBox.UpperLeftY)
    FCurrentBox.UpperLeftY = Vertex.Y;
  if (Vertex.Y > FCurrentBox.LowerRightY)
    FCurrentBox.LowerRightY = Vertex.Y;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Transforms the given coordinate using the provided matrix.
 *
 * @param Matrix The matrix with which to do the transformation. It is a usual 4x4 matrix, although only the upper-left 
 *               2x2 entries are used.
 * @param X The X coordinate.
 * @param Y The Y coordinate.
 * @return The transformed vertex.
 */
TVertex CBoundingBoxComputer::Transform(GLfloat* Matrix, const float& X, const float& Y)
{
  TVertex Result;

  Result.X = X * Matrix[0] + Y * Matrix[4];
  Result.Y = X * Matrix[1] + Y * Matrix[5];

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the current bounding box.
 *
 * @return The current bounding box.
 */
TBoundingBox CBoundingBoxComputer::BoundingBox()
{
  return FCurrentBox;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Takes the new box and merges it with the current bounding box. The new data is not required to be ordered.
 *
 * @param NewBox A new bounding box to merge in.
 */
void CBoundingBoxComputer::Include(GLfloat* Matrix, TBoundingBox* NewBox)
{
  float LLX;
  float URX;
  if (NewBox->UpperLeftX < NewBox->LowerRightX)
  {
    LLX = NewBox->UpperLeftX;
    URX = NewBox->LowerRightX;
  }
  else
  {
    LLX = NewBox->LowerRightX;
    URX = NewBox->UpperLeftX;
  };

  float LLY;
  float URY;
  if (NewBox->UpperLeftY < NewBox->LowerRightY)
  {
    LLY = NewBox->UpperLeftY;
    URY = NewBox->LowerRightY;
  }
  else
  {
    LLY = NewBox->LowerRightY;
    URY = NewBox->UpperLeftY;
  };

  float LLZ;
  float URZ;
  if (NewBox->UpperLeftZ < NewBox->LowerRightZ)
  {
    LLZ = NewBox->UpperLeftZ;
    URZ = NewBox->LowerRightZ;
  }
  else
  {
    LLZ = NewBox->LowerRightZ;
    URZ = NewBox->UpperLeftZ;
  };

  TVertex V;
  if (Matrix != NULL)
    V = Transform(Matrix, LLX, LLY);
  else
  {
    V.X = LLX;
    V.Y = LLY;
  };
  IncludeVertex(V);
  if (Matrix != NULL)
    V = Transform(Matrix, URX, URY);
  else
  {
    V.X = URX;
    V.Y = URY;
  };
  IncludeVertex(V);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Takes the vertex and merges it with the current bounding box.
 *
 * @param Vertex The vertex to be included
 */
void CBoundingBoxComputer::Include(GLfloat* Matrix, const TVertex& Vertex)
{
  TVertex V;
  if (Matrix != NULL)
    V = Transform(Matrix, Vertex.X, Vertex.Y);
  else
    V = Vertex;
  IncludeVertex(V);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Takes the X and Y values and merges them with the current bounding box.
 *
 * @param X The X coordinate of the point to include.
 * @param Y The Y coordinate of the point to include.
 */
void CBoundingBoxComputer::Include(GLfloat* Matrix, const float& X, const float& Y)
{
  TVertex V;
  if (Matrix != NULL)
    V = Transform(Matrix, X, Y);
  else
  {
    V.X = X;
    V.Y = Y;
  };
  IncludeVertex(V);
}

//----------------------------------------------------------------------------------------------------------------------

