/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_view.cpp 
 * @brief Implementation of the view class.
 * 
 */

#include "myx_gc.h"

#include "myx_gc_view.h"

//----------------- CHitResults ----------------------------------------------------------------------------------------

CHitResults::CHitResults(void)
{
  FCurrentEntry = FEntries.end();
}

//----------------------------------------------------------------------------------------------------------------------

CHitResults::~CHitResults(void)
{
  for (THitEntryIterator Iterator = FEntries.begin(); Iterator != FEntries.end(); ++Iterator)
    delete (*Iterator).second;
}

//----------------------------------------------------------------------------------------------------------------------

void CHitResults::AddHit(CFigureInstance* Instance, double Min, double Max)
{
  THitEntry* Entry = new THitEntry;
  Entry->Instance = Instance;
  Entry->ZMax = Max;
  Entry->ZMin = Min;
  FEntries[Instance] = Entry;
}

//----------------------------------------------------------------------------------------------------------------------

int CHitResults::Count(void)
{
  return FEntries.size();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Deletes the given entry from the internal list.
 *
 * @param Instance The entry to be deleted.
 * @return If the entry was in the list then true is returned, otherwise false.
 */
bool CHitResults::Delete(CFigureInstance* Instance)
{
  THitEntryIterator Iterator = FEntries.find(Instance);
  bool Result = Iterator != FEntries.end();
  if (Result)
  {
    if (FCurrentEntry == Iterator)
      ++FCurrentEntry;
    FEntries.erase(Iterator);
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

void CHitResults::DeleteCurrent(void)
{
  if (FCurrentEntry != FEntries.end())
  {
    FEntries.erase(FCurrentEntry);
    ++FCurrentEntry;
  };
}

//----------------------------------------------------------------------------------------------------------------------

bool CHitResults::HasNext(void)
{
  return FCurrentEntry != FEntries.end();
}

//----------------------------------------------------------------------------------------------------------------------

THitEntry* CHitResults::Next(void)
{
  if (FCurrentEntry == FEntries.end())
    return NULL;
  else
  {
    THitEntryIterator Temp = FCurrentEntry;
    ++FCurrentEntry;
    return Temp->second;
  };
}

//----------------------------------------------------------------------------------------------------------------------

void CHitResults::Release(void)
{
  delete this;
}

//----------------------------------------------------------------------------------------------------------------------

void CHitResults::Reset(void)
{
  FCurrentEntry = FEntries.begin();
}

//----------------- CGCView --------------------------------------------------------------------------------------------

CGCView::CGCView(CGenericCanvas* Owner, string Name)
{
  FCanvas = Owner;
  FName = Name;
  FColor[0] = 0;
  FColor[1] = 0;
  FColor[2] = 0;
  FColor[3] = 1;
  FZoomX = 1;
  FZoomY = 1;
  FOffsetX = 0;
  FOffsetY = 0;
  FNearPlane = -100;
  FFarPlane = 100;
  FJitter = 0;
  FIsPicking = false;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Used to set up things that need only to be done once if a view becomes active.
 */
void CGCView::Activate(void)
{
  glClearColor(FColor[0], FColor[1], FColor[2], FColor[3]);

  // TODO: make this dynamically configurable (where applicable).
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glFrontFace(GL_CW);
  glDisable(GL_AUTO_NORMAL);
  glDisable(GL_CULL_FACE);
  glDisable(GL_DITHER);
  glDisable(GL_DEPTH_TEST);

  // glEnable(GL_POINT_SMOOTH);
  // glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glEnable(GL_LINE_SMOOTH);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glEnable(GL_POLYGON_SMOOTH);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glDisable(GL_TEXTURE_2D);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets up the current projection and modelview matrices.
 *
 * @param DoProjection A flag telling if also the project matrix should be set. Useful to avoid unnecessary applications.
 */
void CGCView::ApplyTransformations(bool DoProjection)
{
  if (DoProjection)
  {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-FJitter, FViewport.Width - FJitter, FViewport.Height - FJitter, -FJitter, FNearPlane, FFarPlane);
  };

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslated(FOffsetX, FOffsetY, 0);
  glScaled(FZoomX, FZoomY, 1);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Initiates a repaint of the canvas (and so this view).
 */
void CGCView::Invalidate(void)
{
  if (FCanvas != NULL && FCanvas->CurrentView() == this)
    FCanvas->Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * This is the main paint routine. It must be called by the viewer holding the reference to this canvas (e.g. when a 
 * window must be redrawn).
 */
void CGCView::Render(void)
{
  ApplyTransformations(true);
  for (CLayers::iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    (*Iterator)->Render();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds a layer to the internal list of layers that belong to this view (only if not yet there).
 *
 * @param Layer The layer to add.
 */
void CGCView::AddLayer(CLayer* Layer)
{
  for (CLayers::const_iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    if (*Iterator == Layer)
      return;

  FLayers.push_back(Layer);
  Change(this, GC_CHANGE_VIEW_ADD_LAYER);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Called when something changed in the view.
 *
 * @param AObject An arbitrary object related to the change (usually the view itself).
 * @param Reason An indicator of what has actually changed.
 */
void CGCView::Change(void* AObject, TGCChangeReason Reason)
{
  if (FCanvas != NULL)
  {
    FCanvas->Change(AObject, Reason);
    if (FCanvas->CurrentView() == this)
      FCanvas->Invalidate();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Type neutral helper method to add a new list item to any of the lists in the canvas. This method is meant to be used
 * by scripting engines and things like that because it is not as effective as direct manipulations.
 *
 * @param Path The path to a list property that gets a new entry.
 * @param Name The name of the new list entry. Other properties must be set via CGCView::Property.
 * @returns -1 if Path does not address a list property, otherwise the index of the newly added entry.
 */
int CGCView::AddNewItem(string Path, string Name)
{
  int Result = -1;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
        {
          bool DoAddLayer = true;
          wstring S = Utf8ToUtf16(Name);
          for (CLayers::const_iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
          {
            CLayer* Layer = *Iterator;
            if (Layer->Name() == S)
            {
              // If there is already a layer with that name in this view then do nothing.
              DoAddLayer = false;
              break;
            };
          };

          if (DoAddLayer)
          {
            CLayer* Layer = FCanvas->LayerByName(Name.c_str());
            if (Layer != NULL)
              AddLayer(Layer);
          };
        }
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            if (!Path.empty())
            {
              CLayer* Layer = FLayers[Index];
              Result = Layer->AddNewItem(Path, Name);
            };
          };
        };
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines the child count of the element given by Path.
 *
 * @param The path to the element whose child count is to be determined.
 * @return The number of children or 0 if Path does not point to a list of child objects.
 */
int CGCView::ChildCount(string Path)
{
  int Result = 0;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
          Result = FLayers.size();
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            CLayer* Layer = FLayers[Index];
            Result = Layer->ChildCount(Path);
          };
        };
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the new background color of this view.
 *
 * @param Red The red color component.
 * @param Green The green color component.
 * @param Blue The blue color component.
 * @param Alpha The transparency component.
 */
void CGCView::Color(float Red, float Green, float Blue, float Alpha)
{
  FColor[0] = Red;
  FColor[1] = Green;
  FColor[2] = Blue;
  FColor[3] = Alpha;
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the new background color of this view
 *
 * @param NewColor The new color to use.
 */
void CGCView::Color(GLfloat* NewColor)
{
  FColor[0] = NewColor[0];
  FColor[1] = NewColor[1];
  FColor[2] = NewColor[2];
  FColor[3] = NewColor[3];
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Tells the caller whether this view contains a reference to the given layer.
 *
 * @param Layer The layer to look for.
 * @param True if the layer is referenced in this view otherwise false.
 */
bool CGCView::Contains(CLayer* Layer)
{
  for (CLayers::const_iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    if (*Iterator == Layer)
      return true;

  return false;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Jittering the viewport a little bit sometimes improves display quality (e.g. for thin lines). This function
 * sets this value for this view.
 *
 * @param Value The new jitter value.
 */
void CGCView::Jitter(float Value)
{
  FJitter = Value;
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the current jitter value.
 *
 * @return The current jitter value.
 */
float CGCView::Jitter(void)
{
  return FJitter;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Takes the given coordinates and tries to find a figure that was rendered at this position.
 *
 * @param X The horizontal window (viewer) coordinate.
 * @param Y The vertical coordinate. 
 * @return A hit result class is returned regardless of the actual number of hits. It must be freed by the caller.
 */
CHitResults* CGCView::GetHitTestInfoAt(const int X, const int Y)
{
  FIsPicking = true;

  GLint Viewport[4];
  glGetIntegerv(GL_VIEWPORT, Viewport);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix(X, Y, 4, 4, Viewport);
  glOrtho(-FJitter, FViewport.Width - FJitter, FViewport.Height - FJitter, -FJitter, FNearPlane, FFarPlane);

  GLsizei BufferSize = 500;
  GLuint* Buffer = NULL;
  int Hits = 0;
  glMatrixMode(GL_MODELVIEW);
  do
  {
    Buffer = (GLuint*) realloc(Buffer, BufferSize * sizeof(GLuint));
    glSelectBuffer(BufferSize, Buffer);

    glRenderMode(GL_SELECT);
    glLoadIdentity();

    glInitNames();
    glPushName(0);

    // Render the scene (in select mode nothing is drawn).
    glTranslated(FOffsetX, FOffsetY, 0);
    glScaled(FZoomX, FZoomY, 1);
    for (CLayers::iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    {
      CLayer* Layer = *Iterator;
      if (Layer->Enabled())
        Layer->Render();
    };

    Hits = glRenderMode(GL_RENDER);
    BufferSize <<= 1;
  }
  while (Hits < 0);

  FCanvas->CheckError();
  FIsPicking = false;

  CHitResults* Result = new CHitResults();

  GLuint* Run = Buffer;
  for (int I = 0; I < Hits; I++)
  {
    int Count = *Run++;
    // The depth values are normalized so that the largest value corresponds to 1, while the smallest one is 0.
    // To store this as integer a mapping is applied to make 1 <=> 0xFFFFFFFF.
    double ZMin = (double) *Run++ / 0xFFFFFFFF;
    double ZMax = (double) *Run++ / 0xFFFFFFFF;
    CFigureInstance* Instance = (CFigureInstance*) (*Run++);
    Result->AddHit(Instance, ZMin, ZMax);

    // Usually we have only one entry per hit record.
    // That's the way it works when we only use glLoadName when rendering figures.
    // In order to be error tolerant we skip unused entries accordingly.
    Run += Count - 1;
  };

  free(Buffer);
  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the horizontal offset of the view. The offset is value about which the content of the view is moved.
 *
 * @param Value The new horizontal offset.
 */
void CGCView::OffsetX(float Value)
{
  FOffsetX = Value;
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the current horizontal offset.
 *
 * @return The current horizontal offset.
 */
float CGCView::OffsetX(void)
{
  return FOffsetX;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the new vertical offset.
 *
 * @param Value The new vertical offset.
 */
void CGCView::OffsetY(float Value)
{
  FOffsetY = Value;
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the current vertical offset.
 *
 * @return The curretn vertical offset.
 */
float CGCView::OffsetY(void)
{
  return FOffsetY;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Retrieves the value of the property given by path. The path syntax is must be something like (here expressed as regex)
 * (container)*(property), where container is a slash and the name of a container class (e.g. layers, figures) and
 * property is the name of a simple property of that container.
 *
 * @param Path The path of the property to return.
 * @return A description of the property value and, if the property is simple, the actual value.
 */
TGCVariant CGCView::Property(string Path)
{
  TGCVariant Result;

  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_ROOT:
      {
        switch (GetPropertyID(Path))
        {
          case GC_PROPERTY_NAME:
            {
              Result.Type = GC_VAR_STRING;
              Result.S = FName;
              break;
            };
          case GC_PROPERTY_DESCRIPTION:
            {
              Result.Type = GC_VAR_STRING;
              Result.S = "A view comprising a set of layers.";
              break;
            };
          case GC_PROPERTY_COLOR:
            {
              Result.Type = GC_VAR_STRING;
              Result.S = ColorToString(FColor);
              break;
            };
          case GC_PROPERTY_ZOOMX:
            {
              Result.Type = GC_VAR_FLOAT;
              Result.F = FZoomX;
              break;
            };
          case GC_PROPERTY_ZOOMY:
            {
              Result.Type = GC_VAR_FLOAT;
              Result.F = FZoomY;
              break;
            };
          case GC_PROPERTY_X:
            {
              Result.Type = GC_VAR_FLOAT;
              Result.F = FOffsetX;
              break;
            };
          case GC_PROPERTY_Y:
            {
              Result.Type = GC_VAR_FLOAT;
              Result.F = FOffsetY;
              break;
            };
          case GC_PROPERTY_JITTER:
            {
              Result.Type = GC_VAR_FLOAT;
              Result.F = FJitter;
              break;
            };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
          Result.Type = GC_VAR_LIST;
        else
        {
          // A single entry is requested.
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            if (Path.empty())
              // The entry object is addressed (since no property of it was specified).
              Result.Type = GC_VAR_OBJECT;
            else
            {
              CLayer* Layer = FLayers[Index];
              Result = Layer->Property(Path);
            };
          };
        };
        break;
      };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the value of the given property, which must be a simple property.
 *
 * @param The path of the property.
 * @param Value The new value of the property. Automatic conversion is performed where possible.
 */
void CGCView::Property(string Path, const TGCVariant& Value)
{
  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_ROOT:
      {
        switch (GetPropertyID(Path))
        {
          case GC_PROPERTY_COLOR:
            {
              StringToColor(VariantToString(Value), FColor);
              FColor[3] = 1;
              Change(this, GC_CHANGE_VIEW_PROPERTY);
              break;
            };
          case GC_PROPERTY_ZOOMX:
            {
              FZoomX = VariantToFloat(Value);
              Change(this, GC_CHANGE_VIEW_PROPERTY);
              break;
            };
          case GC_PROPERTY_ZOOMY:
            {
              FZoomY = VariantToFloat(Value);
              Change(this, GC_CHANGE_VIEW_PROPERTY);
              break;
            };
          case GC_PROPERTY_X:
            {
              FOffsetX = VariantToFloat(Value);
              Change(this, GC_CHANGE_VIEW_PROPERTY);
              break;
            };
          case GC_PROPERTY_Y:
            {
              FOffsetY = VariantToFloat(Value);
              Change(this, GC_CHANGE_VIEW_PROPERTY);
              break;
            };
          case GC_PROPERTY_JITTER:
            {
              FJitter = VariantToFloat(Value);
              Change(this, GC_CHANGE_VIEW_PROPERTY);
              break;
            };
        };
        break;
      };
    case GC_CONTAINER_LAYERS:
      {
        if (!Path.empty())
        {
          int Index = GetEntryIndex(Path);
          if ((Index > -1) && (Index < (int) FLayers.size()))
          {
            if (!Path.empty())
            {
              CLayer* Layer = FLayers[Index];
              Layer->Property(Path, Value);
            };
          };
        };
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes an entry from one of our lists (given by path).
 *
 * @param Path The path to the list to modify.
 * @param Index The index of the entry to remove.
 */
void CGCView::RemoveItem(string Path, int Index)
{
  // Extract top container, adjust Path and iterate down if necessary.
  switch (GetTopContainer(Path))
  {
    case GC_CONTAINER_LAYERS:
      {
        if (Path.empty())
        {
          delete FLayers[Index];
          FLayers.erase(FLayers.begin() + Index);
          Change(this, GC_CHANGE_VIEW_REMOVE_LAYER);
        }
        else
        {
          int LayerIndex = GetEntryIndex(Path);
          if ((LayerIndex > -1) && (LayerIndex < (int) FLayers.size()))
          {
            if (!Path.empty())
            {
              CLayer* Layer = FLayers[LayerIndex];
              Layer->RemoveItem(Path, Index);
            };
          };
        };
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 *  Removes the given layer from the list of layers that comprise this view.
 *
 * @param Layer The layer to remove.
 */
void CGCView::RemoveLayer(CLayer* Layer)
{
  for (CLayers::iterator Iterator = FLayers.begin(); Iterator != FLayers.end(); ++Iterator)
    if (*Iterator != Layer)
    {
      FLayers.erase(Iterator);
      Invalidate();
      break;
    };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the new viewport for this view.
 *
 * @param NewViewport The new viewport to be used.
 */
void CGCView::Viewport(const TGCViewport& NewViewport)
{
  FViewport = NewViewport;
  glViewport(FViewport.Left, FViewport.Top, FViewport.Width, FViewport.Height);
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 *  Returns the current viewport.
 *
 * @return The current viewport.
 */
TGCViewport CGCView::Viewport(void)
{
  return FViewport;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets a new horizontal zoom factor.
 *
 * @param Value The new zoom factor.
 */
void CGCView::ZoomX(float Value)
{
  FZoomX = Value;
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the current horizontal zoom factor.
 *
 * @return The current horizontal zomm factor.
 */
float CGCView::ZoomX(void)
{
  return FZoomX;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the new vertical zoom factor.
 *
 * @param Value The new zoom factor.
 */
void CGCView::ZoomY(float Value)
{
  FZoomY = Value;
  Invalidate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Return the current vertical zomm factor.
 *
 * @return The current vertical zoom factor.
 */
float CGCView::ZoomY(void)
{
  return FZoomY;
}

//----------------------------------------------------------------------------------------------------------------------

