/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "myx_grt_gc_bridge.h"

//----------------- General initialization and finalization functions --------------------------------------------------

/**
 * Registers a new GRT-to-GC bridge (as new built-in module) with the given GRT.
 *
 * @param grt The runtime structure with which the bridge should be registered.
 * @return Either MYX_GRT_NO_ERROR if all went fine, otherwise an error code.
 */
MYX_GRT_ERROR myx_register_grt_gc_module(MYX_GRT* Runtime)
{
  CGrtToGcBridge* Bridge = new CGrtToGcBridge(Runtime);

  // Register the new module and keep a reference to it.
  MYX_GRT_MODULE* Module = myx_grt_module_register_builtin(Runtime, &grt_module_grt_gc, Bridge);
  Bridge->Module(Module);

  if (Module != NULL)
    return MYX_GRT_NO_ERROR;
  else                                 
    return MYX_GRT_MODULE_INIT_ERROR;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the GRT-to-GC bridge from the given grt.
 *
 * @param Runtime The runtime structure from which the bridge should be unregistered.
 */
void myx_unregister_grt_gc_module(MYX_GRT* Runtime)
{
  // Look if there is a module with that name.
  MYX_GRT_MODULE* Module = myx_grt_find_module(Runtime, grt_module_grt_gc.name);

  if (Module != NULL)
  {
    CGrtToGcBridge* Bridge = (CGrtToGcBridge*) Module->priv;
    myx_grt_module_unregister_builtin(Runtime, Module);
    delete Bridge;
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Initializes a new mapping between the given canvas and path.
 *
 * @param Runtime The runtime in which the resulting proxy is to be insterted.
 * @param Canvas The canvas that is used for the mapping.
 * @param Path The (absolute) target path for the resulting proxy object.
 */
void myx_grt_gc_bridge_initialize(MYX_GRT *Runtime, CGenericCanvas *Canvas, const char *Path)
{
  // Look if there is a module with that name.
  MYX_GRT_MODULE* Module = myx_grt_find_module(Runtime, grt_module_grt_gc.name);

  if (Module != NULL)
  {
    CGrtToGcBridge* Bridge = (CGrtToGcBridge*) Module->priv;
    Bridge->AddCanvas(Path, Canvas);
    MYX_GRT_VALUE* Canvas = myx_grt_dict_new_proxy(Runtime, Module, Path, "canvas.Canvas");
    myx_grt_dict_item_set_by_path(myx_grt_get_root(Runtime), Path, Canvas);
  };
}

//----------------------------------------------------------------------------------------------------------------------

void myx_grt_gc_bridge_free(MYX_GRT *Runtime, CGenericCanvas *Canvas, const char *Path)
{
  // Look if there is a module with that name.
  MYX_GRT_MODULE* Module = myx_grt_find_module(Runtime, grt_module_grt_gc.name);

  if (Module != NULL)
  {
    CGrtToGcBridge* Bridge = (CGrtToGcBridge*) Module->priv;
    Bridge->RemoveCanvas(Path);
  };
}


//----------------- Module internal functions --------------------------------------------------------------------------

/**
 * Reads and returns the value of a property in one of the registered GCs.
 *
 * @param param The list of parameters.
 * @param data Pointer to the bridge.
 * 
 * @return If parameters are correct and the requrested property exists then its value is returned otherwise NULL.
 */
MYX_GRT_VALUE* grt_gc_value_get(MYX_GRT_VALUE *param, void *data)
{
  MYX_GRT_VALUE *Result = NULL;
  const char* Path = NULL;

  if (myx_grt_value_get_type(param) == MYX_STRING_VALUE)
    Path = myx_grt_value_as_string(param);
  else
    if ((myx_grt_value_get_type(param) == MYX_LIST_VALUE) && (myx_grt_list_item_count(param) > 0))
        Path = myx_grt_list_item_get_as_string(param, 0);

  if (Path != NULL)
  {
    CGrtToGcBridge* Bridge = (CGrtToGcBridge*) data;
    Result = Bridge->GetValue(Path);
  };

  return Result;
}


//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets a new value for the given property, which must be a simply property. Lists and objects are manipulated using
 * other functions.
 *
 * @param param Contains two values. First is a string containing the path to the property to be modified and the second
 *              is the actualy value. Automatic conversion takes place as far as possible.
 * @param data Pointer to the bridge.
 * 
 * @return  The return value if any, otherwise NULL
 */
MYX_GRT_VALUE *grt_gc_value_set(MYX_GRT_VALUE *param, void *data)
{
  MYX_GRT_VALUE *Result = NULL;

  if ((myx_grt_value_get_type(param) == MYX_LIST_VALUE) && (myx_grt_list_item_count(param) > 1))
  {
    const char* Path = myx_grt_list_item_get_as_string(param, 0);
    Result = myx_grt_list_item_get(param, 1);

    if (Path != NULL)
    {
      CGrtToGcBridge* Bridge = (CGrtToGcBridge*) data;
      Bridge->SetValue(Path, Result);
    };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the size of a list whose location is given by param.
 *
 * @param param Contains the path to the list to determine the child count of.
 * @param Pointer to the bridge.
 * 
 * @return The number of child entries in the list.
 */
MYX_GRT_VALUE *grt_gc_list_size(MYX_GRT_VALUE *param, void *data)
{
  MYX_GRT_VALUE *Result = NULL;
  const char* Path = NULL;

  if (myx_grt_value_get_type(param) == MYX_STRING_VALUE)
    Path = myx_grt_value_as_string(param);
  else
    if ((myx_grt_value_get_type(param) == MYX_LIST_VALUE) && (myx_grt_list_item_count(param) > 0))
        Path = myx_grt_list_item_get_as_string(param, 0);

  if (Path != NULL)
  {
    CGrtToGcBridge* Bridge = (CGrtToGcBridge*) data;
    Result = Bridge->GetItemCount(Path);
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the child object of a list given by param as GRT value.
 *
 * @param param Contains the path and index of the list and the required child entry.
 * @param data Pointer to the bridge.
 * 
 * @return The requested child object as GRT value.
 */
MYX_GRT_VALUE *grt_gc_list_item_get(MYX_GRT_VALUE *param, void *data)
{
  MYX_GRT_VALUE *Result = NULL;
  const char* Path = NULL;

  if (myx_grt_value_get_type(param) == MYX_STRING_VALUE)
    Path = myx_grt_value_as_string(param);
  else
    if ((myx_grt_value_get_type(param) == MYX_LIST_VALUE) && (myx_grt_list_item_count(param) > 0))
        Path = myx_grt_list_item_get_as_string(param, 0);

  if (Path != NULL)
  {
    CGrtToGcBridge* Bridge = (CGrtToGcBridge*) data;
    Result = Bridge->GetValue(Path);
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds a new member to the list located at the given path.
 *
 * @param param The list path and the name of the new entry.
 * @param data Pointer to the bridge.
 * 
 * @return The new member.
 */
MYX_GRT_VALUE *grt_gc_list_item_add(MYX_GRT_VALUE *param, void *data)
{
  MYX_GRT_VALUE *Result = NULL;

  if ((myx_grt_value_get_type(param) == MYX_LIST_VALUE) && (myx_grt_list_item_count(param) > 1))
  {
    const char* Path = myx_grt_list_item_get_as_string(param, 0);
    const char* Name = myx_grt_list_item_get_as_string(param, 1);

    if (Path != NULL)
    {
      CGrtToGcBridge* Bridge = (CGrtToGcBridge*) data;
      Result = Bridge->AddListItem(Path, Name);
    };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes an entry from a list given by a path and an index.
 *
 * @param param Two values, one describing the list and one an index in that list.
 * @param data Pointer to the bridge.
 * 
 * @return  The return value is always NULL.
 */
MYX_GRT_VALUE* grt_gc_list_item_remove(MYX_GRT_VALUE *param, void *data)
{
  if ((myx_grt_value_get_type(param) == MYX_LIST_VALUE) && (myx_grt_list_item_count(param) > 1))
  {
    const char* Path = myx_grt_list_item_get_as_string(param, 0);
    MYX_GRT_VALUE* Parameter2 = myx_grt_list_item_get(param, 1);
    int Index = myx_grt_value_as_int(Parameter2);

    if ((Index > -1) && (Path != NULL))
    {
      CGrtToGcBridge* Bridge = (CGrtToGcBridge*) data;
      Bridge->RemoveListItem(Path, Index);
    };
  };

  return NULL;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Grt module function for Bridge implementation
 *
 *   Function to deal with a Grt request for the bridge
 *
 * @param param the list of parameters
 * @param data pointer to Grt
 * 
 * @return  The return value if any, otherwise NULL
 */
MYX_GRT_VALUE *grt_gc_dict_item_get(MYX_GRT_VALUE *param, void *data)
{
  return NULL;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Grt module function for Bridge implementation
 *
 *   Function to deal with a Grt request for the bridge
 *
 * @param param the list of parameters
 * @param data pointer to Grt
 * 
 * @return  The return value if any, otherwise NULL
 */
MYX_GRT_VALUE *grt_gc_dict_item_set(MYX_GRT_VALUE *param, void *data)
{
  return NULL;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Grt module function for Bridge implementation
 *
 *   Function to deal with a Grt request for the bridge
 *
 * @param param the list of parameters
 * @param data pointer to Grt
 * 
 * @return  The return value if any, otherwise NULL
 */
MYX_GRT_VALUE *grt_gc_dict_item_count(MYX_GRT_VALUE *param, void *data)
{
  return NULL;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * @brief Grt module function for Bridge implementation
 *
 *   Function to deal with a Grt request for the bridge
 *
 * @param param the list of parameters
 * @param data pointer to Grt
 * 
 * @return  The return value if any, otherwise NULL
 */
MYX_GRT_VALUE *grt_gc_dict_item_key(MYX_GRT_VALUE *param, void *data)
{
  return NULL;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Extracts the right most path part (separated by a slash character) and returns an ID for it. This path part is then
 * also removed from the given path. If the last part is an integer number (which means it is an index) then
 * the second right most part is used instead.
 *
 * @param Path The path to examine. It must start with a slash to be valid. On return it contains a new 
 *             path where the top level part (if there was one) is removed.
 * @return The last container part in Path.
 */
string GetLastContainerAsString(const string& Path)
{
  string Result;

  if (Path.size() > 0 && Path[0] == '/')
  {
    string::size_type Position = Path.find_last_of('/');

    // The character following the slash is a digit then it is considered as index into its parent collection.
    // In this case extract the parent of the index part instead.
    char Char = Path[Position];
    if (Position < Path.size() - 1)
      Char = Path[Position + 1];
    if (Char < '0' || Char > '9')
      Result = Path.substr(Position);
    else
      if (Position > 0)
      {
        // Digit found, so scan parent part.
        string::size_type NextPosition = Path.find_last_of('/', Position - 1);
        // NextPosition is always valid as we ensured already there is at least one slash at position 0.
        Result = Path.substr(NextPosition, Position - NextPosition);
      };
  };

  return Result;
}

//----------------- CGrtToGcBridge -------------------------------------------------------------------------------------

CGrtToGcBridge::CGrtToGcBridge(MYX_GRT* Runtime)
{
  FRuntime = Runtime;

  FPathToElementMapper["/views"] = "canvas.View";
  FPathToElementMapper["/layers"] = "canvas.Layer";
  FPathToElementMapper["/model"] = "canvas.Model";
  FPathToElementMapper["/scaling"] = "canvas.Scaling";
  FPathToElementMapper["/translation"] = "canvas.Translation";
  FPathToElementMapper["/rotation"] = "canvas.Rotation";
  FPathToElementMapper["/figureInstances"] = "canvas.FigureInstance";
  FPathToElementMapper["/content"] = "canvas.FigureElement";
  FPathToElementMapper["/style"] = "canvas.Style";
  FPathToElementMapper["/styles"] = "canvas.Style";
  FPathToElementMapper["/figure"] = "canvas.Figure";
  FPathToElementMapper["/figures"] = "canvas.Figure";
  FPathToElementMapper["/layouts"] = "canvas.Layout";
  FPathToElementMapper["/groups"] = "canvas.Group";
  FPathToElementMapper["/children"] = "canvas.FigureElement";
  FPathToElementMapper["/caption"] = "canvas.Caption";
  FPathToElementMapper["/constraints"] = "canvas.Constraints";
}

//----------------------------------------------------------------------------------------------------------------------

CGrtToGcBridge::~CGrtToGcBridge(void)
{
  for (CCache::const_iterator Iterator = FValueCache.begin(); Iterator != FValueCache.end(); ++Iterator)
    myx_grt_value_release(Iterator->second);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds a path to canvas mapping to the bridge.
 *
 * @param Path The path under which the canvas can be accessed.
 * @param Canvas The canvas for which a mapping is to be created.
 */
void CGrtToGcBridge::AddCanvas(const char* Path, CGenericCanvas* Canvas)
{
  FCanvasMap[Path] = Canvas;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds a new entry with the given name to the list addressed by Path and returns this new entry to the caller.
 *
 * @param Path The path to list in the canvas hierarchy.
 * @param Name The name of the new list element.
 * @return A new GRT value representing the new list element.
 */
MYX_GRT_VALUE* CGrtToGcBridge::AddListItem(string Path, string Name)
{
  MYX_GRT_VALUE* Result = NULL;

  string ElementPath;
  CGenericCanvas* Canvas = GetCanvas(Path, ElementPath);
  if (Canvas != NULL)
  {
    int Index = Canvas->AddNewItem(ElementPath, Name);

    if (Index > -1)
    {
      // If the returned index is > -1 then a new list item was added. 
      // Use the new index to store a new DICT in the value cache.
      string Container = GetLastContainerAsString(ElementPath);
      if (FPathToElementMapper.find(Container) != FPathToElementMapper.end())
      {
        char Buffer[10]; 
        itoa(Index, Buffer, 10);
        Path += '/';
        Path += Buffer;
        if (FValueCache.find(Path) == FValueCache.end())
        {
          Result = myx_grt_dict_new_proxy(FRuntime, FModule, Path.c_str(), FPathToElementMapper[Container].c_str());
          FValueCache[Path] = Result;
        }
        else
          Result = FValueCache[Path];
      };
    };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Search the canvas map for the given path and returns the associated canvas for it (if any). Search is an incremental
 * process where path parts (separated by a slash) are taken and used as search key. If even with incremental lookup
 * no entry can be found then NULL is returned.
 *
 * @param FullPath A path that can be fully or partially used as a key into the mapper.
 * @param Remainder [out] On return it carries FullPath less the string part used to address the canvas. E.g. when a 
 *                  path like /workbench/model/layer/0/instances/10 is given and a canvas was registered (and found) 
 *                  using /workbench/model as key then /layer/0/instances/10 is returned in Remainder.
 */
CGenericCanvas* CGrtToGcBridge::GetCanvas(string FullPath, string& Remainder)
{
  CGenericCanvas* Result = NULL;

  for (CCanvasMap::const_iterator Iterator = FCanvasMap.begin(); Iterator != FCanvasMap.end(); ++Iterator)
  {
    if (FullPath.compare(0, Iterator->first.size(), Iterator->first) == 0)
    {
      Result = Iterator->second;
      Remainder = FullPath.substr(Iterator->first.size());
      break;
    };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines the number of child elements of the object given by path. This object must be a complex property.
 * 
 * @param Path The absolute path of the object in the GRT.
 * @return A GRT value containing the child count.
 */
MYX_GRT_VALUE* CGrtToGcBridge::GetItemCount(string Path)
{
  MYX_GRT_VALUE* Result = NULL;

  string ElementPath;
  CGenericCanvas* Canvas = GetCanvas(Path, ElementPath);
  if (Canvas != NULL)
  {
    int Value = Canvas->ChildCount(ElementPath);
    Result = myx_grt_value_from_int(Value);
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines values of simple properties or that of lists/dicts. For speed optimization a cache is implemented that
 * keeps MYX_GRT_VALUE structs and only changes their internal value as needed.
 *
 * @param Path The absolute path of the object in the GRT.
 * @return A GRT value containing the property value.
 */
MYX_GRT_VALUE* CGrtToGcBridge::GetValue(string Path)
{
  MYX_GRT_VALUE* Result = NULL;

  string ElementPath;
  CGenericCanvas* Canvas = GetCanvas(Path, ElementPath);
  if (Canvas != NULL)
  {
    TGCVariant Value = Canvas->Property(ElementPath);

    switch (Value.Type)
    {
      case GC_VAR_BOOL:
        {
          if (FValueCache.find(Path) == FValueCache.end())
          {
            Result = myx_grt_value_from_int(Value.B);
            FValueCache[Path] = Result;
          }
          else
            Result = myx_grt_value_change_int(FValueCache[Path], Value.B);
          break;
        };
      case GC_VAR_INT:
        {
          if (FValueCache.find(Path) == FValueCache.end())
          {
            Result = myx_grt_value_from_int(Value.I);
            FValueCache[Path] = Result;
          }
          else
            Result = myx_grt_value_change_int(FValueCache[Path], Value.I);
          break;
        };
      case GC_VAR_FLOAT:
        {
          if (FValueCache.find(Path) == FValueCache.end())
          {
            Result = myx_grt_value_from_real(Value.F);
            FValueCache[Path] = Result;
          }
          else
            Result = myx_grt_value_change_real(FValueCache[Path], Value.F);
          break;
        };
      case GC_VAR_STRING:
        {
          if (FValueCache.find(Path) == FValueCache.end())
          {
            Result = myx_grt_value_from_string(Value.S.c_str());
            FValueCache[Path] = Result;
          }
          else
            Result = myx_grt_value_change_string(FValueCache[Path], Value.S.c_str());
          break;
        };
      case GC_VAR_LIST:
        {
          // It's a list property, that is, a collection of a certain class.
          // Extract container name to know which type to create.
          string Container = GetLastContainerAsString(ElementPath);
          if (FPathToElementMapper.find(Container) != FPathToElementMapper.end())
          {
            if (FValueCache.find(Path) == FValueCache.end())
            {
              Result = myx_grt_list_new_proxy(FRuntime, FModule, Path.c_str(), MYX_DICT_VALUE, 
                FPathToElementMapper[Container].c_str());
              FValueCache[Path] = Result;
            }
            else
              Result = FValueCache[Path];
          };
          break;
        };
      case GC_VAR_OBJECT:
        {
          // Property is a class with subproperties. This corresponds to a DICT in GRT. Create one.
          // Extract container name to know which type to create.
          string Container = GetLastContainerAsString(ElementPath);
          if (FPathToElementMapper.find(Container) != FPathToElementMapper.end())
          {
            if (FValueCache.find(Path) == FValueCache.end())
            {
              Result = myx_grt_dict_new_proxy(FRuntime, FModule, Path.c_str(), FPathToElementMapper[Container].c_str());
              FValueCache[Path] = Result;
            }
            else
              Result = FValueCache[Path];
          };
          break;
        };
    };
  };
  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the module that holds the reference to this bridge.
 *
 * @param AModule The module to keep a reference of.
 */
void CGrtToGcBridge::Module(MYX_GRT_MODULE* AModule)
{
  FModule = AModule;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes a previously established mapping for a certain path from the bridge.
 *
 * @param Path The path under which a mapping was registered.
 */
void CGrtToGcBridge::RemoveCanvas(const char* Path)
{
  FCanvasMap.erase(Path);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes a list entry from a list.
 *
 * @param Path The path to list in the canvas hierarchy.
 * @param Index The index of the member to remove in that list.
 */
void CGrtToGcBridge::RemoveListItem(string Path, int Index)
{
  string ElementPath;
  CGenericCanvas* Canvas = GetCanvas(Path, ElementPath);
  if (Canvas != NULL)
    Canvas->RemoveItem(ElementPath, Index);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the property addressed by Path to a new value .
 *
 * @param Path The path to the property (must be a simply property, no list or dict).
 * @param Value The new value to be set.
 */
void CGrtToGcBridge::SetValue(string Path, MYX_GRT_VALUE* Value)
{
  string ElementPath;
  CGenericCanvas* Canvas = GetCanvas(Path, ElementPath);
  if (Canvas != NULL)
  {
    TGCVariant GCValue;
    switch (Value->type)
    {
      case MYX_INT_VALUE:
        {
          GCValue.Type = GC_VAR_INT;
          GCValue.I = Value->value.i;
          break;
        };
      case MYX_REAL_VALUE:
        {
          GCValue.Type = GC_VAR_FLOAT;
          GCValue.F = (float) Value->value.r;
          break;
        };
      case MYX_STRING_VALUE:
        {
          GCValue.Type = GC_VAR_STRING;
          GCValue.S = Value->value.s;
          break;
        };
    };
    Canvas->Property(ElementPath, GCValue);
  };
}

//----------------------------------------------------------------------------------------------------------------------

