//
//  MQResultSetDataSource.m
//  MySQL QueryBrowser
//
//  Created by Alfredo Kojima on 3/6/05.
//  Copyright 2005 MySQL AB. All rights reserved.
//

#import "MQResultSetDataSource.h"
#import <MySQLToolsCommon/MSparseMatrix.h>

@interface MRSChange : NSObject
{
  MYX_RESULTSET *rs;
@public
  int action_num;
  int *actions;
}
- (id)initWithRS:(MYX_RESULTSET*)resultset;
- (void)addIndex:(int)index;
- (int)valueIndex;
- (BOOL)isDeleted;
@end

@implementation MRSChange
- (id)initWithRS:(MYX_RESULTSET*)resultset
{
  self= [super init];
  if (self)
	rs= resultset;
  return self;
}

- (void)addIndex:(int)index;
{
  action_num++;
  actions= realloc(actions, action_num*sizeof(int));
  actions[action_num-1]= index;
}

- (int)valueIndex
{
  if (action_num == 0)
	return -1;
  if (rs->actions->actions[action_num-1].action == MYX_RSA_DELETE)
	return action_num-2 < 0 ? -1 : actions[action_num-2];
  else
	return actions[action_num-1];
}

- (BOOL)isDeleted
{
  if (action_num == 0)
	return NO;
  if (rs->actions->actions[actions[action_num-1]].action == MYX_RSA_DELETE)
	return YES;
  else
	return NO;
}
@end


@implementation MQResultSetDataSource

- (id)initWithResultSet:(MYX_RESULTSET*)rs
{
  self= [super init];
  if (self)
  {
	_resultset= rs;
 
	_changes= [[MSparseMatrix alloc] init];
	
	_zone= NSCreateZone(256*1024, 64*1024, NO);
  }
  return self;
}

- (MYX_RESULTSET*)resultset
{
  return _resultset;
}

- (void)dealloc
{
  [_changes release];
  NSRecycleZone(_zone);
  myx_query_free_resultset(_resultset);
  [super dealloc];
}

- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
  return _resultset->rows_num + _newRowCount;
}

- (id)tableView:(NSTableView *)aTableView
 objectValueForTableColumn:(NSTableColumn *)aTableColumn
			row:(int)rowIndex
{
  int column= [[aTableColumn identifier] intValue];
  MRSChange *value;
  int index;
  
  value= [_changes objectAtRow:rowIndex
						column:column];
  if (!value || (index= [value valueIndex]) < 0)
  {
	if (rowIndex >= _resultset->rows_num)
	  return @"";
	else
	{
	  MYX_RS_FIELD *row= _resultset->rows[rowIndex].fields;
	  return [[NSString allocWithZone:_zone] initWithBytesNoCopy:row[column].value
														  length:row[column].value_length
														encoding:NSUTF8StringEncoding
													freeWhenDone:NO];
	}
  }
  else
  {
	MYX_RS_ACTION *action= _resultset->actions->actions + index;

	return [[NSString allocWithZone:_zone] initWithBytesNoCopy:action->new_value
														length:action->new_value_length
													  encoding:NSUTF8StringEncoding
												  freeWhenDone:NO];
  }
}


- (void)tableView:(NSTableView *)aTableView 
   setObjectValue:(id)anObject 
   forTableColumn:(NSTableColumn *)aTableColumn 
			  row:(int)rowIndex
{
  int column= [[aTableColumn identifier] intValue];
  MYX_RS_ACTION *action;
  MRSChange *value;
  int aindex;
  NSString *str= (NSString*)anObject;
  
  id ovalue= [self tableView:aTableView
   objectValueForTableColumn:aTableColumn
						 row:rowIndex];
  if (ovalue && [ovalue isEqualTo:anObject])
	return;
  
  action= myx_query_create_action(rowIndex >= _resultset->rows_num ? MYX_RSA_ADD : MYX_RSA_UPDATE,
								  rowIndex,
								  _resultset->columns+column,
								  [str length],
								  [str UTF8String]);

  aindex= myx_query_add_action(_resultset, action);
  
  value= [_changes objectAtRow:rowIndex
						column:column];

  myx_query_free_action(action);

  if (!value)
	value= [[MRSChange alloc] initWithRS:_resultset];

  [value addIndex:aindex];

  [_changes setObject:value
				atRow:rowIndex
			   column:column];
}


- (void)addRow
{
  _newRowCount++;
}

- (void)deleteRow:(int)rowIndex
{
  MYX_RS_ACTION *action;
  MRSChange *value;
  int aindex;
  
  action= myx_query_create_action(MYX_RSA_DELETE,
								  rowIndex,
								  NULL,
								  0, NULL);
  
  aindex= myx_query_add_action(_resultset, action);
  
  myx_query_free_action(action);
  
  value= [_changes objectAtRow:rowIndex
						column:0];
  if (!value)
	value= [[MRSChange alloc] initWithRS:_resultset];
  
  [value addIndex:aindex];
  
  [_changes setObject:value
				atRow:rowIndex
			   column:0];
}


- (void)discardChanges
{
  _newRowCount= 0;
  [_changes removeAllObjects];
  myx_query_discard_actions(_resultset);
}

- (void)prepareForSave
{
  MSparseMatrixRow *row;
  MSparseMatrixColumn *column;
  NSMutableArray *array= [NSMutableArray arrayWithCapacity:32];
  
  row= [_changes firstRow];
  while (row)
  {
	MRSChange *change;
	BOOL deleted= NO;
	change= [_changes objectAtRow:row->row
						   column:0];
	if (change && [change isDeleted])
	{
	  deleted= YES;
	}

	column= [_changes firstColumnInRow:row];
	while (column)
	{
	  unsigned int i;
	  change= column->data;

	  if (deleted)
	  {
		// everything except deleted
		if (column->column == 0)
		{
		  NSAssert(_resultset->actions->actions[change->actions[change->action_num-1]].action == MYX_RSA_DELETE,
				   @"RS Delete Internal Error (bad location for action)");
		  // in column 0, the last action is supposed to be the delete action
		  // for the row
		  for (i= 0; i < change->action_num-1; i++)
			[array addObject:[NSNumber numberWithUnsignedLong:change->actions[i]]];
		}
		else
		{
		  for (i= 0; i < change->action_num; i++)
			[array addObject:[NSNumber numberWithUnsignedLong:change->actions[i]]];
		}
	  }
	  else
	  {
		// all redundant actions
		for (i= 0; i < change->action_num-1; i++)
		{
		  [array addObject:[NSNumber numberWithUnsignedLong:change->actions[i]]];
		}
	  }

	  column= [_changes nextColumnInRow:row after:column];
	}
	row= [_changes nextRowAfter:row];
  }
  
  if ([array count] > 0)
  {
	int i, c;
	[array sortUsingSelector:@selector(compare:)];
	
	c= [array count];
	for (i= c-1; i >= 0; i--)
	{
	  unsigned long index= [[array objectAtIndex:i] unsignedLongValue];
	  
	  myx_query_delete_action(_resultset, index);
	}
  }
}


- (BOOL)isEditable
{
  return _resultset->editable;
}

- (BOOL)editing
{
  return _editEnabled;
}

- (void)setEditing:(BOOL)flag
{
  _editEnabled= flag;
}


- (MQRSStatus)statusOfRow:(int)row
				   column:(int)column
{
  if (row >= _resultset->rows_num)
	return MQAdded;
  else
  {
	MRSChange *change= [_changes objectAtRow:row
									  column:0];
	if (change && [change isDeleted])
	  return MQDeleted;
	else
	{
	  change= [_changes objectAtRow:row
							 column:column];
	  if (change)
		return MQChanged;
	  else
		return MQUnchanged;
	}
  }
}

- (void)commitActions
{
  myx_query_update_resultset(_resultset);

  [self discardChanges];
}

@end
